## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class TriggerTrait(Trait):
    ID      = 'macro'
    WHILE   = 'while'
    UNTIL   = 'until'
    COUNTED = 'counted' # - Always one "do ... while"
    def __init__(self,
                 name            = '',
                 command         = '', # Context menu name
                 key             = '', # Context menu key
                 property        = '', # Enable/Disable
                 watchKeys       = [],
                 actionKeys      = [], # What to do
                 loop            = False,
                 preLoop         = '', # Key
                 postLoop        = '', # Key
                 loopType        = COUNTED, # Loop type
                 whileExpression = '',
                 untilExpression = '',
                 count           = 0,
                 index           = False,
                 indexProperty   = '',
                 indexStart      = '',
                 indexStep       = ''):
        '''Create a layer trait (VASSAL.counter.Trigger)'''
        super(TriggerTrait,self).__init__()
        encWKeys = Trait.encodeKeys(watchKeys, ',')
        encAKeys = Trait.encodeKeys(actionKeys,',')
        self.setType(name            = name,            
                     command         = command,          # Context menu name
                     key             = key,              # Context menu key
                     property        = property,         # Enable/Disable
                     watchKeys       = encWKeys,       
                     actionKeys      = encAKeys,         # What to do
                     loop            = loop,            
                     preLoop         = preLoop,          # Key
                     postLoop        = postLoop,         # Key
                     loopType        = loopType,         # Loop type
                     whileExpression = whileExpression, 
                     untilExpression = untilExpression, 
                     count           = count,           
                     index           = index,           
                     indexProperty   = indexProperty,   
                     indexStart      = indexStart,      
                     indexStep       = indexStep)       
        self.setState(state='')

    def getActionKeys(self):
        return Trait.decodeKeys(self['actionKeys'],',')
    
    def getWatchKeys(self):
        return Trait.decodeKeys(self['watchKeys'],',')
    
    def setActionKeys(self,keys):
        self['actionKeys'] = Trait.encodeKeys(keys,',')
        
    def setWatchKeys(self,keys):
        self['watchKeys'] = Trait.encodeKeys(keys,',')
        
        

Trait.known_traits.append(TriggerTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . head   import num_version
from . features import Features
## END_IMPORT
# ====================================================================
class GBXPieceDef:
    def __init__(self,ar):
        '''Definition of a piece

        FRONT and BACK are tile IDs

        FLAGS is ...
        '''
        with VerboseGuard(f'Reading piece definition'):
            if Features().piece_100:
                n = ar.size()
                self._ids = [ar.iden() for _ in range(n)]
            else:
                self._ids = [ar.word(),ar.word()]
            self._flags = ar.word()


    @property
    def _front(self):
        return self._ids[0]

    @property
    def _back(self):
        return self._ids[1] if len(self._ids) > 1 else 0
    
    def __str__(self):
        return f'Piece: {self._front:04x},{self._back:04x},{self._flags:04x}'
    
# --------------------------------------------------------------------
class GBXPieceSet:
    def __init__(self,ar):
        '''Set of pieces'''
        with VerboseGuard(f'Reading piece set'):
            self._name   = ar.str()
            n            = ar.sub_size()
            self._pieces = [ar.iden() for _ in range(n)]

    def __len__(self):
        return len(self._pieces)

    def __str__(self):
        return (f'{self._name}: '+','.join([str(p) for p in self._pieces]))
    
# --------------------------------------------------------------------
class GBXPieceManager:
    def __init__(self,ar):
        '''Manager of pieces'''
        with VerboseGuard(f'Reading piece manager') as g: 
            self._reserved     = [ar.word() for _ in range(4)]
            g(f'Reserved are {self._reserved}')
            n                  = ar.iden();
            g(f'Will read {n} pieces')
            self._pieces       = [GBXPieceDef(ar) for _ in range(n)]
            n                  = ar.sub_size()
            g(f'Will read {n} sets')
            self._sets         = [GBXPieceSet(ar) for _ in range(n)]

    def __len__(self):
        return len(self._sets)

    def toDict(self,tileManager,strings):
        from math import log10, ceil
        with VerboseGuard(f'Creating dict from piece manager') as gg:
            gg(f'Has {len(self._sets)} and {len(self._pieces)} pieces')
            setDigits   = int(ceil(log10(len(self)+.5)))
            pieceDigits = 1
            for pieceSet in self._sets:
                pieceDigits = max(pieceDigits,
                                  int(ceil(log10(len(pieceSet)+.5))))

            cnt       = 0
            piecesMap = {}
            setList   = []
            ret = {'map': piecesMap,
                   'sets': setList }

            for ips, pieceSet in enumerate(self._sets):
                with VerboseGuard(f'Creating dict from piece set '
                                  f'{pieceSet._name}') as g:
                    setPrefix = f'piece_{ips:0{setDigits}d}'
                    idList    = []
                    setDict   = { 'description': pieceSet._name.strip(),
                                  'pieces':      idList }
                    setList.append(setDict)
                    
                    for ipc, pieceID in enumerate(pieceSet._pieces):
                        
                        piecePrefix = f'{setPrefix}_{ipc:0{pieceDigits}d}'
                        pieceDef    = self._pieces[pieceID]
                        tmpStr      = strings._id2str.get(pieceID,'')
                        pieceDesc   = tmpStr.replace('\r','').replace('\n',', ').replace('/','\\/')
                        pieceDict   = {}
                        if pieceDesc != '':
                            pieceDict['description'] = pieceDesc
                        cnt += 1
                        # pieceList.append(pieceDict)
                        idList   .append(pieceID)
                        
                        # print(f'{pd}  => "{tr}"')
                        for tileId,which in zip([pieceDef._front,
                                                 pieceDef._back],
                                                ['front',
                                                 'back']):                    
                            img = tileManager.image(tileId)
                    
                            if img is None:
                                continue
                    
                            sav     = f'{piecePrefix}_{which}.png'
                            setname = pieceSet._name.strip()\
                                .replace('\n',' ')\
                                .replace('\r',' ')\
                                .replace('/','\\/')
                            gg(f'Set name, escaped: "{setname}"')
                            pieceDict[which] = {
                                'image':     img,
                                'filename':  sav,
                                'set':       setname }
                        piecesMap[pieceID] = pieceDict
                        g(f'{pieceID}: {pieceDict}')

            gg(f'{list(piecesMap.keys())}')
            return ret
    
        
    def __str__(self):
        return ('Piece manager:\n'
                +f'Reserved: {self._reserved}\n'
                +f'# pieces: {len(self._pieces)}\n  '
                +'\n  '.join([str(p) for p in self._pieces])+'\n'
                +f'# piece sets: {len(self._sets)}\n  '
                +'\n  '.join([str(p) for p in self._sets])
                )

# --------------------------------------------------------------------
class GSNPieceEntry:
    def __init__(self,ar,vers,i):
        '''Manager of pieces'''
        with VerboseGuard(f'Reading piece # {i:3d} ({vers//256}.{vers%256})'):
            self._side   = ar.byte()
            self._facing = ar.word()
            self._owner  = ar.word() if vers < num_version(3,10) else ar.dword()

    def __str__(self):
        return f'Piece: {self._side}, {self._facing:3d}, {self._owner:08x}'
    
# --------------------------------------------------------------------
class GSNPieceTable:
    def __init__(self,ar,vers):
        '''Manager of pieces'''
        with VerboseGuard(f'Reading piece table'):
            self._reserved     = [ar.word() for _ in range(4)] 
            n                  = ar.word()#sub_size()
            if Features().piece_100:
                dummy          = ar.word();
            self._pieces       = [GSNPieceEntry(ar,vers,i) for i in range(n)]
            
    def __str__(self):
        pl = '\n    '.join([str(s) for s in self._pieces])
        return (f'Piece table: {self._reserved} {len(self._pieces)}'
                f'\n    {pl}\n')
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
## END_IMPORT

class GSNWindow:
    def __init__(self,ar):
        with VerboseGuard('Reading window state') as g:
            self._code   = ar.word()
            self._user   = ar.word()
            self._board  = ar.iden()
            self._state  = [ar.dword() for _ in range(10)]
            n            = ar.size()
            g(f'Read {self._code} {self._user} {self._board} {self._state}')
            g(f'Reading {n} bytes at {ar.tell()}')
            self._buffer = ar.read(n)

    def __str__(self):
        return (f'code={self._code:04x} '
                f'user={self._user:04x} '
                f'board={self._board:04x} '
                f'buffer={len(self._buffer)}')
        
class GSNWindows:
    def __init__(self,ar):
        with VerboseGuard(f'Reading window states') as g: 
            self._savePositions = ar.word()
            self._enable        = ar.byte()
            n                   = ar.size() if self._enable else 0
            g(f'Save position: {self._savePositions}, '
              f'enable {self._enable} '
              f'n={n}')
            self._states        = [GSNWindow(ar) for _ in range(n)]

    def __str__(self):
        pl = '\n    '.join([str(s) for s in self._states])
        return f'Windows states ({len(self._states)})\n    {pl}\n'
    

#
# EOF
#



## BEGIN_IMPORTS
from .. common import VerboseGuard, Verbose
from .. vassal.buildfile import BuildFile
from .. vassal.documentation import Documentation
from .. vassal.traits import *
from .. vassal.base import *
from .. vassal.moduledata import ModuleData
from .. vassal.exporter import Exporter
from .. vassal.mapelements import LayerControl
## END_IMPORTS

class CbExporter(Exporter):
    def __init__(self,
                 d,
                 title         = None,
                 version       = None,
                 rules         = None,
                 tutorial      = None,
                 visible       = False,
                 main_board    = None,
                 map_regex     = None,
                 vassalVersion = '3.6',
                 do_scenario   = False):
        self._d             = d
        self._title         = title
        self._version       = version
        self._rules         = rules
        self._tutorial      = tutorial
        self._visible       = visible
        self._vassalVersion = vassalVersion
        self._flipped       = False
        self._mainBoard     = main_board
        self._mapRegex      = map_regex
        self._do_scenario   = do_scenario
        self._description   = d['description']
        self._placements    = {}

    # ----------------------------------------------------------------
    def createModuleData(self):
        '''Create the `moduleData` file in the module
        '''
        with VerboseGuard(f'Creating module data'):
            self._moduleData = ModuleData()
            data             = self._moduleData.addData()
            data.addVersion      (version     = self._version)
            data.addVASSALVersion(version     = self._vassalVersion)
            data.addName         (name        = self._title)
            data.addDescription  (description = self._description)
            data.addDateSaved    ()

    # ----------------------------------------------------------------
    def addMoreDocumentation(self,doc):
        pass
    
    # ----------------------------------------------------------------
    def addDocumentation(self):
        '''Add documentation to the module.  This includes rules,
        key-bindings, and about elements.
        '''
        with VerboseGuard('Adding documentation') as v:
            doc = self._game.addDocumentation()
            if self._rules is not None:
                self._vmod.addExternalFile(self._rules,'rules.pdf')
                doc.addBrowserPDFFile(title   = 'Show rules',
                                      pdfFile = 'rules.pdf')
            
            if self._tutorial is not None:
                self._vmod.addExternalFile(self._tutorial,'tutorial.vlog')
                doc.addTutorial(name            = 'Tutorial',
                                logfile         = 'tutorial.vlog',
                                launchOnStartup = True)
            
            keys = [
                ['Alt-A',	'-',	'Show the charts panel'],
                ['Alt-B',	'-',	'Show the OOBs'],
                ['Alt-C',	'-',	'Show the counters panel'],
                ['Alt-E',	'-',	'Show the eliminated units'],
                ['Alt-I',	'-',	'Show/refresh inventory window'],
                ['Alt-M',	'-',	'Show map'],
                ['Alt-T',	'-',	'Increase turn track'],
                ['Alt-Shift-T', '-',	'Decrease turn track'],
                ['Alt-6',	'-',	'Roll the dice'],
                ['Ctrl-D',	'Board,Counter','Delete counters'],
                ['Ctrl-E',	'Board,Counter','Eliminate counters'],
                ['Ctrl-F',	'Board,Counter','Flip counters'],
                ['Ctrl-M',	'Board,Counter','Toggle "moved" markers'],
                ['Ctrl-O',	'Board',	'Hide/show counters'],
                ['Ctrl-R',	'Board,Counter','Restore unit'],
                ['Ctrl-T',	'Board,Counter','Toggle move trail'],
                ['Ctrl-+',	'Board',	'Zoom in'],
                ['Ctrl--',	'Board',	'Zoom out'],
                ['Ctrl-=',	'Board',	'Select zoom'],
                ['Ctrl-Shift-O',	'Board','Show overview map'],
                ['&larr;,&rarr;,&uarr;&darr;','Board',
                 'Scroll board left, right, up, down (slowly)'],
                ['PnUp,PnDn','Board',	'Scroll board up/down (fast)'],
                ['Ctrl-PnUp,Ctrl-PnDn','Board',
                 'Scroll board left/right (fast)'],
                ['Mouse-scroll up/down',	'Board',
                 'Scroll board up//down'],
                ['Shift-Mouse-scroll up/down','Board',
                 'Scroll board right/leftown'],
                ['Ctrl-Mouse-scroll up/down','Board','Zoom board out/in'],
                ['Mouse-2',	'Board',	'Centre on mouse']]
            
            
            self._vmod.addFile('help/keys.html',
                               Documentation.createKeyHelp(
                                   keys,
                                   title=self._d['title'],
                                   version=self._d['version']))
            doc.addHelpFile(title='Key bindings',fileName='help/keys.html')

            self.addMoreDocumentation(doc)

    # ----------------------------------------------------------------
    def saveImages(self,d,vmod):
        for serial,board in d['boards'].items():
            self.saveSVG(board,vmod)

        for piece in d['pieces']['map'].values():
            if 'back' in piece: self._flipped = True
            
            for which in ['front','back']:
                if which not in piece:
                    continue
            
                side     = piece[which]
                self.savePNG(side,vmod)
            
        for mark in d['marks']['map'].values():
            self.savePNG(mark,vmod)
            
    # ----------------------------------------------------------------
    def saveSVG(self,d,vmod):

        with VerboseGuard(f'Saving rendered SVG image {d["filename"]}') as vg:
            from wand.image import Image as WandImage
            from pathlib import Path

            path          = Path(d['filename'])
            filename      = path.with_suffix('.png')
            image         = d['image']
            size          = d['size']
            d['filename'] = filename
            svg           = image.encode()
            svgout        = path.with_suffix('.svg')
            pngout        = path.with_suffix('.png')
            vmod.addFile('images/'+str(svgout),svg)
            vg(f'Creating WandImage')
            
            # We write the SVG to file, because Wand seems to have
            # some issues reading from a bytes.  The same applies for
            # the command line
            if vg:
                vg(f'=== Write the SVG image to {svgout} ===')
                with open(svgout,'w') as out:
                    out.write(image)
                
            with WandImage() as wimage:
                # Reading the bytes directly fails.  It doesn't work
                # from reading BytesIO stream either.
                vg(f'Reading in SVG code')                
                wimage.read(filename=svgout,
                            # blob=image,#.encode(),
                            width=size[0],
                            height=size[1],
                            format='svg')

                # Make a PNG
                vg(f'Making blob')
                png = wimage.make_blob('png')

                # Write to the VMOD 
                vg(f'Write to VMOD')
                vmod.addFile('images/'+str(pngout),png)
                vg(f'Done writing SVG as PNG to VMOD')

    # ----------------------------------------------------------------
    def savePNG(self,d,vmod):
        with VerboseGuard(f'Saving PNG image {d["filename"]}'):
            from io import BytesIO
        
            filename  = d['filename']
            img       = d['image']
            d['size'] = img.size
            stream   = BytesIO()
            img.save(stream,format='PNG')
            vmod.addFile('images/'+filename,stream.getvalue())

    # ----------------------------------------------------------------
    def addPrototypes(self,sides):
        with VerboseGuard('Adding prototypes') as v:
            protos = self._game.addPrototypes()
            traits = [ReportTrait(key('E'),key('R'),key('M')),
                      TrailTrait(),
                      RotateTrait(),
                      DeleteTrait(),
                      MarkTrait(name='PieceLayer',value='Pieces'),
                      SendtoTrait(mapName     = 'DeadMap',
                                  boardName   = f'Pool',
                                  name        = 'Eliminate',
                                  key         = key('E'),
                                  restoreName = 'Restore',
                                  restoreKey  = key('R'),
                                  description = 'Eliminate unit'),
                      BasicTrait()]
            
            protos.addPrototype(name        = f'Piece prototype',
                                description = f'Prototype for pieces',
                                traits      = traits)
            traits = [DeleteTrait(),
                      RotateTrait(),
                      MarkTrait(name='PieceLayer',value='Markers'),
                      BasicTrait()]
            protos.addPrototype(name        = f'Marker prototype',
                                description = f'Prototype for markers',
                                traits      = traits)

            for side in sides:
                v(f'Making prototype trait for faction: "{side}"')
                escside = side.replace('/','\\/')
                traits = [MarkTrait(name='Faction',value=escside),
                          BasicTrait()]
                protos.addPrototype(name        = f'{side} prototype',
                                    description = f'Prototype for {side}',
                                    traits      = traits)
            
    # ----------------------------------------------------------------
    def sanitise(self,txt):
        for special in ['/',',',';','\\','\n','\t','\r']:
            txt = txt.replace(special,' ')
        return txt

    # ----------------------------------------------------------------
    def getToAdd(self,master,mapping,children):
        toAdd = {}
        for iden in children:
            if iden not in master:
                print(f'ID={iden} not found')
                continue
            if iden in mapping:
                # Already added
                continue

            toAdd[iden] = master[iden]

        return toAdd

    # ----------------------------------------------------------------
    def addPieceContainer(self,container,name):
        #panel = container.addPanel(entryName = name,
        #                           fixed = False,
        #                           vert  = True)
        #return panel
        panel = container.addPanel(entryName = name, fixed = False)
        plist = panel.addList(entryName = f'{name}',
                              width=300,
                              height=300,
                              divider=150)
        return plist
        
    # ----------------------------------------------------------------
    def addPieces(self,container,pieces,*pieceSets):
        from pathlib import Path 
        with VerboseGuard('Adding pieces') as v:
            # Store a map from pieceID to PieceSlot
            self._pieceMap = {}
            for sets in pieceSets:
                for set in sets:
                    toAdd = self.getToAdd(pieces,self._pieceMap,set['pieces'])
                    if len(toAdd) <= 0:
                        # Nothing to add from this set
                        continue 

                    name  = set.get('name',set.get('description',None))
                    if name is None:
                        print(f'No name for set: {set}')
                        continue
                    
                    plist = self.addPieceContainer(container, name)
                    
                    for pieceID, piece in toAdd.items():
                        v(f'Make piece with faction: "{name}"')
                        escname = name.replace('/','\\/')
                        front  = piece['front']['filename']
                        size   = piece['front']['size']
                        back   = piece.get('back',{}).get('filename',None)
                        traits = [MovedTrait(xoff=size[0]//2,
                                             yoff=-size[1]//2),
                                  MarkTrait(name='Faction',value=escname),
                                  MarkTrait(name='PieceLayer',value='Pieces'),
                                  PrototypeTrait(name='Piece prototype')]
                
                        if back is not None:
                            traits.extend([
                                LayerTrait(images = [front, back],
                                           newNames = ['','Reduced +'],
                                           activateName = '',
                                           decreaseName = '',
                                           increaseName = 'Flip',
                                           increaseKey  = key('F'),
                                           decreaseKey  = '',
                                           name         = 'Step'),
                                ReportTrait(key('F'))])
                        if 'description' in piece:
                            desc = self.sanitise(piece['description'])
                            traits.append(MarkTrait(name='description',
                                                    value=desc))
                
                        pname = Path(front).stem.replace('_front','')
                        gpid  = self._game.nextPieceSlotId()
                        traits.extend([BasicTrait(name     = pname,
                                                  filename = front,
                                                  gpid     = gpid)])
                                      
                        ps = plist.addPieceSlot(entryName = pname,
                                                gpid      = gpid,
                                                traits    = traits,
                                                width     = size[0],
                                                height    = size[1])

                        # print(f'Piece: {pieceID}')
                        self._pieceMap[pieceID] = ps

    # ----------------------------------------------------------------
    def addMarks(self,container,marks,*markSets):
        from pathlib import Path 
        with VerboseGuard('Adding marks') as v:
            # Store a map from markID to PieceSlot
            self._markMap = {}
            for sets in markSets:
                for set in sets:
                    toAdd = self.getToAdd(marks,self._markMap,set['marks'])
                    if len(toAdd) <= 0:
                        # Nothing to add from this set
                        continue 

                    name  = set.get('name',set.get('description',None))
                    if name is None:
                        print(f'No name for set: {set}')
                        continue
                    
                    mlist = self.addPieceContainer(container, name)
                    
                    for markID, mark in toAdd.items():
                        file   = mark['filename']
                        size   = mark['size']
                        traits = [PrototypeTrait(name='Mark prototype'),
                                  MarkTrait(name='PieceLayer',
                                            value='Markers')]
                        if 'description' in mark:
                            desc = self.sanitise(mark['description'])
                            traits.append(MarkTrait(name='description',
                                                    value=desc))
            
                        mname = Path(file).stem
                        gpid  = self._game.nextPieceSlotId()
                        traits.extend([BasicTrait(name     = mname,
                                                  filename = file,
                                                  gpid     = gpid)])
                                      
                        ps = mlist.addPieceSlot(entryName = mname,
                                                gpid      = gpid,
                                                traits    = traits,
                                                width     = size[0],
                                                height    = size[1])

                        # print(f'Mark:   {markID}')
                        self._markMap[markID] = ps
            
    # ----------------------------------------------------------------
    def addBoards(self,boards):
        with VerboseGuard(f'Adding boards '
                          f'{[n["gboard"]["name"] for n in  boards]}') as v:
            first = True
            names = []
            for board in boards:
                names.append(self.addBoard(first,**board))
                first = False

            filtered = [n for n in names
                        if n != '' and n != self._mainBoard]
            v(f'Filtered set of maps: {filtered}')
            if len(filtered) > 1:
                self._game.addMenu(description = 'Maps',
                                   text        = 'Maps',
                                   tooltip     = 'Show or hide maps',
                                   icon        = '/images/map.gif',
                                   menuItems   = filtered)

    # ----------------------------------------------------------------
    def addMapDefaults(self,map):
            # map.addHidePiecesButton()
            # Basics
            map.addStackMetrics()
            map.addImageSaver()
            map.addTextSaver()
            map.addForwardToChatter()     
            map.addMenuDisplayer()        
            map.addMapCenterer()          
            map.addStackExpander()        
            map.addCounterDetailViewer()
            map.addPieceMover()           
            map.addKeyBufferer()
            map.addForwardKeys()
            map.addScroller()
            map.addSelectionHighlighters()
            map.addHighlightLastMoved()   
            map.addZoomer()               
        
    # ----------------------------------------------------------------
    def addMapLayers(self,map):
        layerDesc = {'Pieces':   {'t': 'Pieces', 'i': ''},
                     'Markers':  {'t': 'Markers','i': ''}}
        layers = map.addLayers(property = 'PieceLayer',
                               layerOrder = layerDesc)
        for layerTitle, layerData in layerDesc.items():
            layers.addControl(name    = layerTitle,
                              tooltip = f'Toggle display of {layerTitle}',
                              text    = f'Toggle {layerData["t"]}',
                              icon    = layerData['i'],
                              layers  = [layerTitle])
        layers.addControl(name = 'Show all',
                          tooltip = 'Show all',
                          text    = 'Show all',
                          command = LayerControl.ENABLE,
                          layers  = list(layerDesc.keys()))
        map.addMenu(description = 'Toggle layers',
                    text        = '',
                    tooltip     = 'Toggle display of layers',
                    icon        = '/images/inventory.gif',
                    menuItems   = ([f'Toggle {ln["t"]}'
                                        for ln in layerDesc.values()]
                                       +['Show all']))         
    # ----------------------------------------------------------------
    def addBoard(self,first,gboard,color=0xFF0000,pieces=None,indicators=None):
        with VerboseGuard(f'Adding board {gboard["name"]}') as v:
            mapName = gboard['name']
            map     = self._game.addMap(
                mapName    = mapName,
                buttonName = '' if first else mapName,
                hotkey     = '' if first else key('M',ALT),
                launch     = not first
            )
            self.addMapDefaults(map)
            self.addMapLayers(map)
            map.addGlobalMap()
            
            map.addMassKey(name='Eliminate',
                           buttonHotkey = key('E'),
                           hotkey       = key('E'),
                           icon         = '/icons/16x16/edit-undo.png',
                           tooltip      = 'Eliminate selected units')
            map.addMassKey(name='Delete',
                           buttonHotkey = key('D'),
                           hotkey       = key('D'),
                           icon         = '/icons/16x16/no.png',
                           tooltip      = 'Delete selected units')
            if self._flipped:
                map.addMassKey(name='Flip',
                               buttonHotkey = key('F'),
                               hotkey       = key('F'),
                               icon         = '/images/Undo16.gif',
                               tooltip      = 'Flip selected units')
                

            size   = gboard['size']
            picker = map.addBoardPicker()
            board  = picker.addBoard(name   = gboard['name'],
                                     image  = gboard['filename'],
                                     width  = size[0],
                                     height = size[1])
            zoned  = board.addZonedGrid()
            zoned.addHighlighter()
            zone   = zoned.addZone(name = 'Full',
                                   useParentGrid = False,
                                   path=(f'{0},{0};' +
                                         f'{size[0]},{0};' +
                                         f'{size[0]},{size[1]};' +
                                         f'{0},{size[1]}'))

            self.addGrid(zone,gboard,color)
            self.addAtStart(map,pieces)
            self.addAtStart(map,indicators)

            return mapName
            
    # ----------------------------------------------------------------
    def addCharts(self,gcharts):
        if len(gcharts) <= 0:
            return
        
        with VerboseGuard(f'Adding Charts '
                          f'{[n["gboard"]["name"] for n in gcharts]}') as v:
            charts = \
                self._game.addChartWindow(name='Charts',
                                          hotkey = key('A',ALT),
                                          description = 'Charts',
                                          text        = '',
                                          icon       = '/images/chart.gif',
                                          tooltip     = 'Show/hide Charts')
            tabs = charts.addTabs(entryName='Charts')

            for chart in gcharts:
                widget = tabs.addMapWidget(entryName=chart['gboard']['name'])
                self.addChart(widget,**chart)
                
    # ----------------------------------------------------------------
    def addChart(self,widget,gboard,color=0xFF0000,pieces=None,indicators=None):
        with VerboseGuard(f'Adding Chart {gboard["name"]}') as v:
            map = widget.addWidgetMap(mapName   = gboard['name'],
                                      markMoved = 'Never',
                                      hotkey    = '')
            self.addMapDefaults(map)
            self.addMapLayers(map)
            
            size   = gboard['size']
            picker = map.addPicker()
            board  = picker.addBoard(name  = gboard['name'],
                                     image = gboard['filename'])
            zoned  = board.addZonedGrid()
            zoned.addHighlighter()
            zone   = zoned.addZone(name = 'Full',
                                   useParentGrid = False,
                                   path=(f'{0},{0};' +
                                         f'{size[0]},{0};' +
                                         f'{size[0]},{size[1]};' +
                                         f'{0},{size[1]}'))
            
            self.addGrid(zone,gboard,color)
            self.addAtStart(map,pieces)
            self.addAtStart(map,indicators)
            
    # ----------------------------------------------------------------
    def getNumParam(self,col,gboard,geom):
        rcol = (col >> 16) & 0xFF
        gcol = (col >> 8)  & 0xFF
        bcol = (col >> 0)  & 0XFF
        fid  = gboard['numbering']['order'].lower()
        sid  = 'h' if fid=='v' else 'v'
        side = 'sideways' in geom['shape']
        hex  = 'hexagon'  in geom['shape']
        d = {'first':      fid.upper(),
             'color':      rgb(rcol,gcol,bcol),
             'visible':    self._visible,
             'vOff':       gboard['rows']['offset']+1,
             'vDescend':   gboard['rows']['inverted'],
             'hOff':       gboard['columns']['offset']+1, # CB start at 1
             'vOff':       gboard['columns']['inverted']+1, # VSL start at 0
             'hLeading':   1 if gboard['numbering']['padding'] else 0,
             'vLeading':   1 if gboard['numbering']['padding'] else 0,
             'stagger':    geom['stagger'] == 'in',
             f'{fid}Type': gboard['numbering']['first'],
             f'{sid}Type': 'N'
             }
        # Coord 0 -> A in VSL, Coord 1 -> A in CB
        d[f'{fid}Off'] += 0 if gboard['numbering']['first'] != 'A' else -1
        return d
            
    # ----------------------------------------------------------------
    def addGrid(self,zone,gboard,col=0x000000):
        geom = gboard.get('cells',{}).get('geometry',None)
        if geom is None:
            return

        if geom['shape'] == 'rectangle':
            self.addRegularGrid(zone,gboard,geom,col)
        elif 'brick' in geom['shape']:
            self.addRegionGrid(zone,gboard,geom,col)
        elif 'hexagon' in geom['shape']:
            self.addHexGrid(zone,gboard,geom,col)

        

    # ----------------------------------------------------------------
    def addHexGrid(self,zone,gboard,geom,col=0x000000):
        from math import sin, pi
        with VerboseGuard(f'Adding hex grid') as v:
            size = geom['size']
            side = 'sideways' in geom['shape']
            npar = self.getNumParam(col,gboard,geom)
            dx   = int(size[0] * sin(pi/3) ** 2)
            dy   = size[1]
            x0   = size[0]//2
            y0   = size[1]//2 + (0 if geom['stagger']=='out' else dy//2)
            if side:
                dy            = size[0]
                dx            = dy * sin(pi/3)
                x0, y0        = y0, x0
                npar['first'] = 'V' if npar['first'] == 'H' else 'H'
                
            v(f'Size = {size}, side = {side} dx={dx} dy={dy}')
            grid = zone.addHexGrid(dx           = dx,
                                   dy           = dy,
                                   x0           = x0,
                                   y0           = y0,
                                   sideways     = side,
                                   visible      = self._visible,
                                   color        = npar['color'],
                                   edgesLegal   = True,
                                   cornersLegal = True,
                                   snapTo       = True
                                   #gboard['snap']['enable']
                                   )
            grid.addNumbering(**npar)

    # ----------------------------------------------------------------
    def addRegularGrid(self,zone,gboard,geom,col=0x000000):
        size = geom['size']
        npar = self.getNumParam(col,gboard,geom)
        grid = zone.addSquareGrid(dx      = size[0],
                                  dy      = size[1],
                                  x0      = size[0]//2,
                                  y0      = size[1]//2,
                                  visible = self._visible,
                                  color   = npar['color'],
                                  snapTo  = gboard['snap']['enable'])
        grid.addNumbering(**npar)
                          
                          

    # ----------------------------------------------------------------
    def addRegionGrid(self,zone,gboard,geom):
        grid = zone.addRegionGrid(snapto  = True,
                                  visible = self._visible)
        
        for row in gboard['cells']['list']:
            for cell in row:
                name = f'r{cell["row"]}_{cell["column"]}'
                pixel = cell['pixel']
                grid.addRegion(name      = name,
                               originx   = pixel[0],
                               originy   = pixel[1],
                               alsoPiece = False)
    # ----------------------------------------------------------------
    def addAtStart(self,map,pieces):
        with VerboseGuard(f'Adding at-start to {map["mapName"]}') as g:
            toAdd = {}
            mapName = map['mapName']
            
            for piece in pieces:
                pmap = None
                if piece['type'] == 'Piece':
                    pmap = self._pieceMap
                elif piece['type'] == 'Mark':
                    pmap = self._markMap
                else:
                    continue
            
            
                iden = piece['id']
                pm   = pmap.get(iden,None)
                #print(piece,pm)
                if pm is None:
                    print(f'Cannot fine {piece["type"]} {iden}')
                    print(list(pmap.keys()))
                    continue

                grid = tuple(piece['grid'])
                x = piece['pixel']['x']
                y = piece['pixel']['y']
                if grid not in toAdd:
                    toAdd[grid] = {'center': (x,y),
                                   'pieces': [] }
                toAdd[grid]['pieces'].append(pm)

            if self._do_scenario:
                mapped = {
                    a['center']: a['pieces'] for a in toAdd.values()}

                if mapName not in self._placements:
                    self._placements[mapName] = {}

                mapPlacements = self._placements[mapName]
                
                # We have to do this carefully, because we may have
                # already added something to the placements for this
                # map and specific locations.
                for coord, pieces in mapped.items():
                    if coord in mapPlacements:
                        mapPlacements[coord].extend(pieces)
                    else:
                        mapPlacements[coord] = pieces 

            else:
                for grid, dpieces in toAdd.items():
                    center  = dpieces['center']
                    name    = f'{grid[0]:02d}{grid[1]:02d}'
                    atstart = map.addAtStart(name=name,
                                             useGridLocation=False,
                                             owningBoard=mapName,
                                             x = center[0],
                                             y = center[1])
                    atstart.addPieces(*dpieces['pieces'])
        
    # ----------------------------------------------------------------
    def addDeadMap(self,sides):
        '''Add a "Dead Map" element to the module 
        '''
        name = 'DeadMap'
        with VerboseGuard(f'Adding board {name}') as v:
            map    = self._game.addMap(mapName       = name,
                                       buttonName    = '',
                                       markMoved     = 'Never',
                                       launch        = True,
                                       icon          = '/images/playerAway.gif',
                                       allowMultiple = True,
                                       hotkey        = key('E',ALT))
            # Basics
            map.addStackMetrics()
            map.addImageSaver()
            map.addTextSaver()
            map.addForwardToChatter()     
            map.addMenuDisplayer()        
            map.addMapCenterer()          
            map.addStackExpander()        
            map.addPieceMover()           
            map.addKeyBufferer()          
            map.addSelectionHighlighters()
            map.addHighlightLastMoved()   
            map.addZoomer()               
            
            map.addMassKey(name='Restore',
                           buttonHotkey = key('R'),
                           hotkey       = key('R'),
                           icon         = '/images/Undo16.gif',
                           tooltip      = 'Restore selected units')

            if sides is None or len(sides) <= 0:
                sides = ['A']
            picker = map.addBoardPicker()
            picker.addSetup(maxColumns=len(sides),mapName=name,
                            boardNames=[s+' pool' for s in sides])
            
            for i, s in enumerate(sides):
                v(f'Adding {s} pool')
                color        = [0,0,0,64]
                color[i % 3] = 255
                w            = 400
                h            = 400
                c            = rgba(*color)
                board        = picker.addBoard(name   = f'{s} pool',
                                               image  = '',
                                               width  = w,
                                               height = h,
                                               color  = c)
                 
# ====================================================================
class GBXExporter(CbExporter):
    def __init__(self,
                 d,
                 title         = None,
                 version       = None,
                 rules         = None,
                 tutorial      = None,
                 visible       = False,
                 main_board    = None,
                 map_regex     = None,
                 vassalVersion = '3.6',
                 do_scenario   = False):
        super(GBXExporter,self).__init__(d,
                                         title         = title,
                                         version       = version,
                                         rules         = rules,
                                         tutorial      = tutorial,
                                         visible       = visible,
                                         main_board    = main_board,
                                         map_regex     = map_regex,
                                         vassalVersion = vassalVersion,
                                         do_scenario   = do_scenario)

    # ----------------------------------------------------------------
    def setup(self):
        self.saveImages(self._d,self._vmod)
        self._title       = self._d['title']
        self._version     = self._d['version']
        self._description = self._d['description']
        
    # ----------------------------------------------------------------
    def createBuildFile(self,
                        ignores = ['all',
                                   'common',
                                   'marker',
                                   'markers',
                                   ' ']):
        with VerboseGuard(f'Creating build file') as v:
            self._build = BuildFile() # 'buildFile.xml')
            self._game  = self._build.addGame(name        = self._title,
                                              version     = self._version,
                                              description = self._description)
            self.addDocumentation()
            self._game.addBasicCommandEncoder()


# ====================================================================
class GSNExporter(CbExporter):
    def __init__(self,
                 d,
                 title         = None,
                 version       = None,
                 rules         = None,
                 tutorial      = None,
                 visible       = False,
                 main_board    = None,
                 map_regex     = None,
                 vassalVersion = '3.6',
                 do_scenario   = False):
        super().__init__(d,
                         title         = title,
                         version       = version,
                         rules         = rules,
                         tutorial      = tutorial,
                         visible       = visible,
                         main_board    = main_board,
                         map_regex     = map_regex,
                         vassalVersion = vassalVersion,
                         do_scenario   = do_scenario)

    # ----------------------------------------------------------------
    def setup(self):
        with VerboseGuard('Setup') as v:
            from pathlib import Path
                
            self.saveImages(self._d['gamebox'],self._vmod)
            self.setupTitle()

    def setupTitle(self):
        with VerboseGuard('Setup titles and such') as v:
            v(f'Before: '
              f'title={self._title}, '
              f'version={self._version}, '
              f'description={self._description}')
            v(f'Game box: '
              f'title={self._d["gamebox"]["title"]}, '
              f'version={self._d["gamebox"]["version"]}, '
              f'description={self._d["gamebox"]["description"]}')
            v(f'Scenario: '
              f'title={self._d["title"]}, '
              f'version={self._d["version"]}, '
              f'description={self._d["description"]}')
            
            if self._title is None or self._title == '':
                # If we are making a scenario, then do not take the
                # title from the scenario file
                self._title = None if self._do_scenario else self._d['title']
                if self._title is None or self._title == '':
                    self._title = self._d['gamebox']['title']
                    if self._title is None or self._title == '':
                        from random import choice
                        from string import ascii_lowercase
                        self._title = Path(self._vmod.fileName()).stem
                        # self._title = ''.join([choice(ascii_lowercase)
                        #                        for _ in range(32)])
            if not self._do_scenario:
                self._d['title'] = self._title
                
            if self._version is None or self._version == '':
                # If we are doing a scenario, do not take the version
                # from the scenario file.
                self._version = (None if self._do_scenario else
                                 self._d['version'])
                if self._version is None or self._version == '':
                    self._version = self._d['gamebox']['version']
                    if self._version is None or self._version == '':
                        self._version = '0.0'
            if not self._do_scenario:
                self._d['version'] = self._version

            if self._do_scenario:
                # Take overall description form game box
                self._description = self._d['gamebox']['description']
            v(f'After: '
              f'title={self._title}, '
              f'version={self._version}, '
              f'description={self._description}')
        
    # ----------------------------------------------------------------
    def createBuildFile(self,
                        ignores = '(.*markers?|all|commons|[ ]+)'):
        '''Create the XML buildFile.xml
 
        Parameters
        ----------
        ignores : str
            Regular expression to match ignored categories for factions
            determination. Python's re.fullmatch is applied to this
            regular exression against chit categories.  If the pattern
            is matched, then the chit is not considered to belong to a
            faction.
        '''
        with VerboseGuard(f'Creating build file') as v:                    
            desc = None if self._do_scenario else self._d['description']
            if desc is None or desc == '':
                desc = self._d['gamebox']['description']
            if len(desc) > 32:
                desc = desc[:32]
            self._build = BuildFile() # 'buildFile.xml')
            self._game  = self._build.addGame(name        = self._title,
                                              version     = self._version,
                                              description = desc)

            self._sides  = self._d['players']
            self._sides  = [p.replace(' player','').replace(' Player','')
                            for p in self._sides]

            if not self._sides:
                self._sides = [s['description'] for s in
                               self._d['gamebox']['pieces']['sets']]

            v(f'Sides: {self._sides}')
            
            v(f'Adding documentation')
            self.addDocumentation()
            
            v(f'Adding command encoder')
            self._game.addBasicCommandEncoder()

            v(f'Adding Global options')
            go = self._game.addGlobalOptions(
                autoReport         = 'Use Preferences Setting',
                centerOnMove       = 'Use Preferences Setting',
                nonOwnerUnmaskable = 'Use Preferences Setting',
                playerIdFormat     = '$playerName$')
            go.addOption(name='undoHotKey',value=key('Z'))
            
            v(f'Adding player roster {self._sides}')
            roster = self._game.addPlayerRoster()
            for side in self._sides:
                roster.addSide(side)
            
            v(f'Adding global properties')
            glob = self._game.addGlobalProperties()
            glob.addProperty(name='TurnTracker.defaultDocked',
                             initialValue=True)
            
            v(f'Adding notes')
            self._game.addNotes()
            
            v(f'Adding turn track')
            turns = self._game.addTurnTrack(name='Turn',
                                            counter={
                                                'property': 'Turn',
                                                'phases': {
                                                    'property': 'Phase',
                                                    'names': self._sides } })

            
            self.addPrototypes(self._sides)
            self.addPiecesMarks()
            self.addBoards()
            self.addCharts()
            self.addDeadMap(self._sides)
            self.addScenario()
            

    #
    def addMoreDocumentation(self,doc):
        desc = f'''<html><body>
        <h1>Module</h1>
        <p>
        This module was created from CyberBoard Scenario
        {self._d["title"]} version {self._d["version"]} and game box
        {self._d["gamebox"]["title"]} version {self._d["gamebox"]["version"]} '
        'by the Python script <code>gsnexport.py</code> available from
        </p>
        <pre>
        htps://gitlab.com/wargames_tex/pywargame
        </pre>
        <h1>Game</h1>
        <table>
          <tr><td>Title:</td><td>{self._d["gamebox"]["title"]}</td></tr>
          <tr><td>Version:</td><td>{self._d["gamebox"]["version"]}</td></tr>
          <tr><td>Author:</td><td>{self._d["gamebox"]["author"]}</td></tr>
          <tr><td>Description:</td><td>{self._d["gamebox"]["description"]}</td></tr>
        </table>
        <h1>Scenario</h1>
        <table>
          <tr><td>Title:</td><td>{self._d["title"]}</td></tr>
          <tr><td>Version:</td><td>{self._d["version"]}</td></tr>
          <tr><td>Author:</td><td>{self._d["author"]}</td></tr>
          <tr><td>Description:</td><td>{self._d["description"]}</td></tr>
        </table>
        </body></html>;'''
        self._vmod.addFile('help/description.html',desc)
        doc.addHelpFile(title='Description',fileName='help/description.html')
        

    # ----------------------------------------------------------------
    def addPiecesMarks(self):
        with VerboseGuard(f'Adding pieces and marks') as v:
            window = self._game.addPieceWindow(name   = 'Counters',
                                               hotkey = key('C',ALT))
            combo  = window.addCombo(entryName='Counters')
            self.addPieces(combo,
                           self._d['gamebox']['pieces']['map'],
                           self._d['trays'],
                           self._d['gamebox']['pieces']['sets'])
            self.addMarks(combo,
                          self._d['gamebox']['marks']['map'],
                          self._d['gamebox']['marks']['sets'])
    # ----------------------------------------------------------------
    def boardMatch(self,name):
        if self._mapRegex is None:
            return False
        
        from re import match

        ret = match(self._mapRegex, name) is not None

        return ret
    
    # ----------------------------------------------------------------
    def addBoards(self):
        with VerboseGuard(f'Adding boards') as v:
            toAdd = []
            for serial, sboard in self._d['boards'].items():
                board = self._d['gamebox']['boards'].get(serial,None)
                if board is None:
                    continue
                # print('BOARD (gsn)',board)
                #
                # If not set to start map in GBX, and name does not
                # the user specified main board and board name does
                # not match the board regular expression, then
                # continue on (the board will be made chart instead).
                if not sboard['onload'] and \
                   board['name'] != self._mainBoard and \
                   not self.boardMatch(board['name']):
                    continue

                # Here, we add pieces and indicators as at-start on
                # the map.  Perhaps we should _not_ do that, and write
                # a scenario file instead.
                d = {'gboard':      board,
                     'color':       sboard['grid']['color'],
                     'pieces':      sboard['pieces'],
                     'indicators':  sboard['indicators'] }
                if board['name'] == self._mainBoard:
                    # If this is the main board, as selected by user,
                    # then put first in the queue.
                    toAdd.insert(0,d)
                else:
                    toAdd.append(d)

            super(GSNExporter,self).addBoards(toAdd)
            
    # ----------------------------------------------------------------
    def addCharts(self):
        with VerboseGuard(f'Adding Charts') as v:
            toAdd = [] 
            for serial, sboard in self._d['boards'].items():
                board = self._d['gamebox']['boards'].get(serial,None)
                if board is None:
                    continue
                # print('BOARD (gbx)',board)
                #
                # If the board is set to start at load on GBX, or the
                # name of the board matches the main board name set by
                # user, or board name matched the user-defined board
                # regular expression, the skip this board (was added
                # previously as a board).
                if sboard['onload'] or \
                   board['name'] == self._mainBoard or \
                   self.boardMatch(board['name']):
                    continue

                # Here, we add pieces and indicators as at-start on
                # the map.  Perhaps we should _not_ do that, and write
                # a scenario file instead.
                d = {'gboard':      board,
                     'color':       sboard['grid']['color'],
                     'pieces':      sboard['pieces'],
                     'indicators':  sboard['indicators'] }
                if board['name'] == self._mainBoard:
                    # If this is the main board, as selected by user,
                    # then put first in the queue.
                    #
                    # Should this be done for charts? Probably not. 
                    toAdd.insert(0,d)
                else:
                    toAdd.append(d)
            
            super(GSNExporter,self).addCharts(toAdd)
            
    # ----------------------------------------------------------------
    def addScenario(self):
        if not self._do_scenario or len(self._placements) <= 0:
            # Nothing to do
            return

        from pathlib import Path 

        menu = self._game.addPredefinedSetup(name = 'Setups',
                                             isMenu = True)

        menu.addPredefinedSetup(name = 'No setup',
                                file = '',
                                useFile = False,
                                description = 'No initial setup')

        with VerboseGuard(f'Adding a scenario {self._d["title"]}') as v:
            vsav    = VSav(self._build,self._vmod)
            save    = vsav.addSaveFile()

            # from pprint import pprint
            # pprint(self._placements)
            
            for mapName,placements in self._placements.items():
                
                save.addNoGrid(mapName, placements)

            # print('\n'.join(save.getLines()))
            
            savfile = self._d['title'].replace(' ','_')+'.vsav'
            vsav.run(savename    = savfile,
                     description = self._d['description'])

            menu.addPredefinedSetup(name=self._d['title'],
                                    useFile=True,
                                    file=savfile,
                                    description=self._d['description'])

        
# ====================================================================
            
#
# EOF
#

            
            




## BEGIN_IMPORT
from .. common import VerboseGuard
## END_IMPORT

# ====================================================================
class GBXImage:
    def __init__(self,ar,transparent=None,save=None):
        '''A DIB image stored in GameBox'''
        with VerboseGuard('Reading an image') as g:
            size = ar.dword()
                
            if size & 0x80000000: # Compressed
                from zlib import decompress
                g(f'Read compressed image')
                
                size       &= 0x80000000  # Remove flag
                compSize   =  ar.dword() # Compressed size
                compressed =  ar.read(compSize) # Compressed
                buffer     = decompress(compressed,bufsize=size)
                #assert len(buffer) == size, \
                #    f'Failed to decompress to expected {size}, ' \
                #    f'got {len(buffer)}'
                    
            else:
                buffer  = ar.read(size)
            
            from PIL import Image as PILImage
            from io import BytesIO
            from numpy import asarray, where, uint8
            
            img = PILImage.open(BytesIO(buffer)).convert('RGBA')
            
            # If transparancy is defined, clean up the image 
            if transparent is None:
                self._img = img
            else:
                g(f'Making #{transparent:06x} transparent')
                dat         = asarray(img)
                tr          = (transparent >> 16) & 0xFF
                tg          = (transparent >> 8)  & 0xFF
                tb          = (transparent >> 0)  & 0xFF
                dat2        = dat.copy()
                dat2[:,:,3] = (255*(dat[:,:,:3]!=[tb,tg,tr]).any(axis=2))\
                    .astype(uint8)

                #if (dat[:,:,3] == dat2[:,:,3]).all():
                #    print(f'Transparency operation seemed to have no effect '
                #          f'for image')
                
                self._img  = PILImage.fromarray(dat2)
            
            if save is None:
                return
            
            self._img.save(save)

    @classmethod
    def b64encode(cls,img):
        '''Encode image as a base64 data URL'''
        from io import BytesIO
        from base64 import b64encode

        if img is None:
            return None
        
        buffered = BytesIO()
        img.save(buffered,format='PNG')
        data = b64encode(buffered.getvalue())
        if not isinstance(data,str):
            data = data.decode()

        return 'data:image/png;base64,'+data

#
# EOF
#



class Geometry:
    RECTANGLE        = 0
    HORIZONTAL_BRICK = 1
    VERTICAL_BRICK   = 2
    HEXAGON          = 3
    SIDEWAYS_HEXAGON = 4
    STAGGER_OUT      = 0
    STAGGER_IN       = 1
    TYPES = {
        RECTANGLE       : 'rectangle',
        HORIZONTAL_BRICK: 'horizontal brick',
        VERTICAL_BRICK  : 'vertical brick',
        HEXAGON         : 'hexagon',
        SIDEWAYS_HEXAGON: 'sideways hexagon'
    }
    STAGGERS = {
        STAGGER_OUT: 'out',
        STAGGER_IN:  'in'
    }
    SKETCHES = {
        RECTANGLE       : r'''
|---|---|
|---|---|''',
        HORIZONTAL_BRICK: r'''
|---|---|
  |---|''',
        VERTICAL_BRICK  : r'''
|--|  |
|  |--|
|--|  |''',
        HEXAGON         : r''' _
/ \
\_/'''        ,
        SIDEWAYS_HEXAGON: r'''/\
||
\/'''
    }

    def __init__(self,nrows,ncols,tpe,stagger,w,h):
        self._nrows   = nrows
        self._ncols   = ncols
        self._type    = tpe
        self._stagger = stagger
        self._width   = w
        self._height  = h
        self._dx      = w
        self._dy      = h

        if self._type == self.HEXAGON:
            self._dx = int(0.75 * self._dx)
        elif self._type == self.SIDEWAYS_HEXAGON:
            self._dy = int(0.75 * self._dy)

    def toPixel(self,row,col):
        x = col * self._dx
        y = row * self._dy
        if self._type == self.RECTANGLE: # No offset for rectangles
            return x,y
        elif self._type in [self.HORIZONTAL_BRICK,self.SIDEWAYS_HEXAGON]:
            x += self._dx//2 if (row % 2) != self._stagger else 0
        elif self._type in [self.VERTICAL_BRICK,self.HEXAGON]:
            y += self._dy//2 if (col % 2) != self._stagger else 0
        return x,y

    def toGrid(self,x,y):
        col = x / self._dx
        row = y / self._dy
        if self._type in [self.HORIZONTAL_BRICK,self.SIDEWAYS_HEXAGON]:
            col -= .5 if (int(row) % 2) != self._stagger else 0
        if self._type in [self.VERTICAL_BRICK,self.HEXAGON]:
            row -= .5  if (int(col) % 2) != self._stagger else 0
        
        return int(row), int(col)

    def boardSize(self,nrows,ncols):
        w = ncols * self._dx
        h = nrows * self._dy

        if self._type in [2,3]:
            h += self._dy // 2
        if self._type in [1,4]:
            w += self._dx // 2
        if self._type == 3:
            w += self._dx // 3
        if self._type == 4:
            h += self._dy // 3

        return w+1,h+1

    def test(self,stepX=None,stepY=None):
        from numpy import linspace
        
        width, height = self.boardSize(self._nrows,self._ncols)
        print(f'Test: {self._type},{self._stagger}')
        print(f'{self._width}x{self._height} ({self._dx},{self._dy})')
        print(f'      {self.TYPES[self._type]} {self.STAGGERS[self._stagger]}')
        print(self.SKETCHES[self._type])

        for row in range(self._nrows):
            for col in range(self._ncols):
                x, y   = self.toPixel(row,col)
                cx, cy = x + self._width // 2, y + self._height // 2
                r, c   = self.toGrid(cx,cy)
                print(f'== {row:3d},{col:3d} -> {cx:6d},{cy:6d} -> {r:3d},{c:3d}')
                assert r == row, f'Rows (from {row} to {r}) inconsistent'
                assert c == col, f'Columns (from {col} to {c}) inconsistent'

                if row >= 3 or col >= 3:
                    continue 
                for dx in linspace(-self._width/4,self._width/4,5):
                    idx = int(dx)
                    for dy in linspace(-self._height/4,self._height/4,5):
                        idy = int(dy)
                        xx = cx + idx
                        yy = cy + idy
                        r, c = self.toGrid(xx,yy)
                        if r != row or c != col:
                            print(f'{row:3d},{col:3d} -> {cx:6d}+{idx:3d}={xx:6d},{cy:6d}+{idy:3d}={yy:6d} -> {r:3d},{c:3d}')
            
                    


def testit(tpe,stagger):
    w = 100
    h = 100
    if tpe == 3:
        w = 80
        h = 68
    elif tpe == 4:
        w = 68
        h = 80

    geo = Geometry(5,5,tpe,stagger,w,h)
    geo.test()


def testall():
    for t in [Geometry.RECTANGLE,
              Geometry.HORIZONTAL_BRICK,
              Geometry.VERTICAL_BRICK,
              Geometry.HEXAGON,
              Geometry.SIDEWAYS_HEXAGON]:
        for s in [Geometry.STAGGER_IN,
                  Geometry.STAGGER_OUT]:
            testit(t,s)



if __name__ == '__main__':
    testall()
    
    
                
    
        
        
        



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class ReportTrait (Trait):
    ID = 'report'
    def __init__(self,
                 *keys,
                 nosuppress = True,
                 description = '',
                 report      = '$location$: $newPieceName$ $menuCommand$ *',
                 cyclekeys   = '',
                 cyclereps   = ''):
        '''Create a report trait (VASSAL.counters.ReportActon)'''
        super(ReportTrait,self).__init__()
        esckeys = ','.join([k.replace(',',r'\,') for k in keys])
        esccycl = ','.join([k.replace(',',r'\,') for k in cyclekeys])
        escreps = ','.join([k.replace(',',r'\,') for k in cyclereps])
        
        self.setType(keys         = esckeys,
                     report       = report,
                     cycleKeys    = esccycl,
                     cycleReports = escreps,
                     description  = description,
                     nosuppress   = nosuppress)
        self.setState(cycle = -1)

Trait.known_traits.append(ReportTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
# Inset
# obs;88,130;ag_hide_1.png;Reveal;I;?;sides:Argentine;Peek;;true;;
# obs;88,130;ag_hide_1.png;Reveal;I;?;side:Argentine;;;true;;
#
# Peek
# obs;88,130;ag_hide_1.png;Reveal;P89,130;?;sides:Argentine;Peek;;true;;
#
# Image
#
class MaskTrait(Trait):
    ID = 'obs'
    INSET = 'I'
    BACKGROUND = 'B'
    PEEK = 'P'
    IMAGE = 'G'
    INSET2 = '2'
    PLAYER = 'player:'
    SIDE = 'side:'
    SIDES = 'sides:'
    def __init__(self,
                 keyCommand   = '',
                 imageName    = '',
                 hideCommand  = '',
                 displayStyle = '',
                 peekKey      = '',
                 ownerImage   = '',
                 maskedName   = '?',
                 access       = '',#?
                 peekCommand  = '',
                 description  = '',
                 autoPeek     = True,
                 dealKey      = '',
                 dealExpr     = ''):
        '''Create a masking trait'''
        super(MaskTrait,self).__init__()
        disp = displayStyle
        if displayStyle == self.PEEK:
            disp += peekKey
        elif displayStyle == self.IMAGE:
            disp += ownerImage
            
        acc = self.PLAYER
        if isinstance(access,list):
            acc = self.SIDES + ':'.join(access)
        elif access.startswith('player'):
            acc = self.PLAYER
        elif access.startswith('side'):
            acc = self.SIDE
                
        self.setType(keyCommand   = keyCommand,
                     imageImage   = imageName,
                     hideCommand  = hideCommand,
                     displayStyle = disp,
                     maskedName   = maskedName,
                     access       = acc, # ?
                     peekCommand  = peekCommand,
                     description  = description,
                     autoPeek     = autoPeek,
                     dealKey      = dealKey,
                     dealExpr     = dealExpr)
        self.setState(value='null')

    @classmethod
    def peekDisplay(cls,key):#Encoded key
        return cls.PEEK + key
    
    @classmethod
    def peekImage(cls,ownerImage):
        return cls.IMAGE + ownerImage

    @classmethod
    def sides(cls,*names):
        return cls.SIDES+':'.join(names)

Trait.known_traits.append(MaskTrait)

#
# EOF
#



#!/usr/bin/env python
## BEGIN_IMPORTS
from pywargame.vassal import Merger
from pywargame.common import Verbose
## END_IMPORTS

# ====================================================================        
def mergeMain():
    from argparse import ArgumentParser, FileType

    ap = ArgumentParser(description='Merge two modules or extensions')
    ap.add_argument('input',type=FileType('r'),help='Input files',nargs='+')
    ap.add_argument('-o','--output',type=FileType('w'),
                    help='Output file',nargs='?',default='Merged.vmod')
    ap.add_argument('-V','--verbose',action='store_true',
                    help='Be verbose')
    ap.add_argument('-p','--patch',type=FileType('r'),
                    help='Python patch script to execute after merging')
    ap.add_argument('-O','--overwrite',action='store_true',
                    help='Overwrite elements, attributes, files, '
                    'etc. with later specified content')
    ap.add_argument('-K','--keep',action='store_false',dest='overwrite',
                    help='Keep first elements, attributes, files, etc.')
    ap.add_argument('-S','--assume-same',action='store_true',
                    help='Assume all modules are same game')


    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)
    
    if len(args.input) < 2:
        raise RuntimeError('At least two inputs must be given')

    outname = args.output.name
    args.output.close()

    merger = Merger(outname,*args.input)

    try:
        merger.run(patch       = args.patch,
                   overwrite   = args.overwrite,
                   assume_same = args.assume_same)
    except:
        from pathlib import Path

        op = Path(outname)
        op.unlink(missing_ok=True)
        
        raise
    
        
# ====================================================================        
if __name__ == '__main__':
    mergeMain()
    

#
#
#



## BEGIN_IMPORTS
from . base import DomInspect
from pywargame.common import VerboseGuard
## END_IMPORTS

# --------------------------------------------------------------------
class ZTDiceHand(DomInspect):
    
    def __init__(self,elem,type=6,count=1):
        with VerboseGuard(f'Got a dice hand') as v:
            if elem is None:
                self._type = type
                self._dice = [
                    {'color': 'black',
                     'pips':  'white',
                     'texture': ''}] * count
                return 
            
            self._type  = int(self._get_attr(elem,'type','D6')
                              .lower().replace('d',''))
            self._dice  = []

            # v(f'{elem.getElementsByTagName("dice")} child dice')
            for dice in elem.getElementsByTagName('dice'):
                count    = int(self._get_attr(dice,'count',1))
                colour   = self._parse_hex(self._get_attr(dice,'color',
                                                          0xFFFFFF))
                pips     = self._parse_hex(self._get_attr(dice,'pips',
                                                          0x000000))
                texture  = self._get_attr(dice,'texture-file')

                #v(f'Adding another dice: "{colour}" "{pips}" "{count}"')
                self._dice.extend([{
                    'color': colour,
                    'pips':  pips,
                    'texture': texture}]*count)
                
            # v(f'{len(self._dice)} dice of type d{self._type}')


    def __str__(self):
        return (f' Dice-hand: {self._type}'+
                ('\n' if len(self._dice) > 0 else '')+
                '\n'.join([f'  {d["color"]},{d["pips"]},{d["texture"]}'
                           for d in self._dice]))
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class SubMenuTrait(Trait):
    ID = 'submenu'
    def __init__(self,
                 subMenu     = '',  # Title
                 keys        = [],  # Keys
                 description = ''):
        '''Create a sub menu (VASSAL.counters.SubMenu)'''
        self.setType(subMenu     = subMenu,   # CLONEKEY
                     keys        = ','.join([k.replace(',',r'\,')
                                             for k in keys]),
                     description = description)
        self.setState() # PROPERTY COUNT (followed by [; KEY; VALUE]+)
    def setKeys(self,keys):
        '''Set the keys'''
        self['keys'] = ','.join([k.replace(',',r'\,') for k in keys])
        
Trait.known_traits.append(SubMenuTrait)

#
# EOF
#



#!/usr/bin/env python

from . verboseguard import VerboseGuard
from . verbose import Verbose

def inner():
    with VerboseGuard('Entering inner') as g:
        g('Message from inner')
        
def test():
    with VerboseGuard('Entering test'):
        inner()
    

if __name__ == '__main__':
    from argparse import ArgumentParser

    ap = ArgumentParser('test verbose')
    ap.add_argument('-v','--verbose',help='Be verbose',action='store_true')

    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)
    
    test()



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
## END_IMPORT

# --------------------------------------------------------------------
HEX_WIDTH = 88.50779626676963
HEX_HEIGHT = 102.2
RECT_WIDTH  = 80
RECT_HEIGHT = 80
# --------------------------------------------------------------------
class BaseGrid(Element):
    def __init__(self,zone,tag,node=None,
                 color        = rgb(0,0,0),
                 cornersLegal = False,
                 dotsVisible  = False,
                 dx           = HEX_WIDTH,  # Meaning seems reversed!
                 dy           = HEX_HEIGHT,
                 edgesLegal   = False,
                 sideways     = False,
                 snapTo       = True,
                 visible      = True,
                 x0           = 0,
                 y0           = 32):
        super(BaseGrid,self).__init__(zone,tag,node=node,
                                      color        = color,
                                      cornersLegal = cornersLegal,
                                      dotsVisible  = dotsVisible,
                                      dx           = dx,
                                      dy           = dy,
                                      edgesLegal   = edgesLegal,
                                      sideways     = sideways,
                                      snapTo       = snapTo,
                                      visible      = visible,
                                      x0           = x0,
                                      y0           = y0)
    def getZone(self):
        ## BEGIN_IMPORT
        from . zone import Zone
        ## END_IMPORT
        z = self.getParent(Zone)
        return z
    def getZonedGrid(self):
        z = self.getZone()
        if z is not None:
            return z.getZonedGrid()
        return None
    def getBoard(self):
        z = self.getZonedGrid()
        if z is not None:
            return z.getBoard()
        return self.getParent(Board)
    def getPicker(self):
        z = self.getBoard()
        if z is not None:
            return z.getPicker()
        return None
    def getMap(self):
        b = self.getPicker()
        if b is not None:
            return b.getMap()
        return None
    def getNumbering(self):
        pass 
    def getLocation(self,loc):
        numbering = self.getNumbering()
        if numbering is None or len(numbering) < 1:
            return None

        return numbering[0].getLocation(loc)
    
# --------------------------------------------------------------------
class BaseNumbering(Element):
    def __init__(self,grid,tag,node=None,
                 color                = rgb(255,0,0),
                 first                = 'H',
                 fontSize             = 24,
                 hDescend             = False,
                 hDrawOff             = 0,
                 hLeading             = 1,
                 hOff                 = 0,
                 hType                = 'A',
                 locationFormat       = '$gridLocation$',
                 rotateText           = 0,
                 sep                  = '',
                 stagger              = True,
                 vDescend             = False,
                 vDrawOff             = 32,
                 vLeading             = 0,
                 vOff                 = 0,
                 vType                = 'N',
                 visible              = True):
        super(BaseNumbering,self).__init__(grid,tag,node=node,
                                           color           = color,
                                           first           = first,
                                           fontSize        = fontSize,
                                           hDescend        = hDescend,
                                           hDrawOff        = hDrawOff,
                                           hLeading        = hLeading,
                                           hOff            = hOff,
                                           hType           = hType,
                                           locationFormat  = locationFormat,
                                           rotateText      = rotateText,
                                           sep             = sep,
                                           stagger         = stagger,
                                           vDescend        = vDescend,
                                           vDrawOff        = vDrawOff,
                                           vLeading        = vLeading,
                                           vOff            = vOff,
                                           vType           = vType,
                                           visible         = visible)
    def getGrid(self): return getParent(BaseGrid)

    def _getMatcher(self,tpe,leading):
        if tpe == 'A':
            return \
                '-?(?:A+|B+|C+|D+|E+|F+|G+|H+|I+|'  + \
                'J+|K+|L+|M+|N+|O+|P+|Q+|R+|S+|T+|' + \
                'U+|V+|W+|X+|Y+|Z+)'

        return f'-?[0-9]{{{int(leading)+1},}}'

    def _getIndex(self,name,tpe):
        if tpe == 'A':
            negative = name.startswith('-')
            if negative:
                name = name[1:]

            value = 0
            for num,let in enumerate(name):
                if not let.isupper():
                    continue
                if num < len(name) - 1:
                    value += 26
                else:
                    value += ord(let)-ord('A')

            if negative:
                value *= -1

            return value

        return int(name)

    def _getCenter(self,col,row):
        '''Convert col and row index to picture coordinates'''
        print('Dummy GetCenter')
        pass
    
    def getLocation(self,loc):
        '''Get picture coordinates from grid location'''
        from re import match

        first   = self['first']
        vType   = self['vType']
        hType   = self['hType']
        vOff    = int(self['vOff'])
        hOff    = int(self['hOff'])
        sep     = self['sep']
        colPat  = self._getMatcher(hType,self['hLeading'])
        rowPat  = self._getMatcher(vType,self['vLeading'])
        patts   = ((colPat,rowPat) if first == 'H' else (rowPat,colPat))
        colGrp  = 1 if first == 'H' else 2
        rowGrp  = 2 if first == 'H' else 1
        patt    = sep.join([f'({p})' for p in patts])
        matched = match(patt,loc)
        if not matched:
            return None

        rowStr  = matched[rowGrp]
        colStr  = matched[colGrp]
        rowNum  = self._getIndex(rowStr,vType)
        colNum  = self._getIndex(colStr,hType)

        ret = self._getCenter(colNum-hOff, rowNum-vOff);
        #print(f'Get location of "{loc}" -> {rowNum},{colNum} -> {ret}')
        return ret
       
    
# --------------------------------------------------------------------
class HexGrid(BaseGrid):
    TAG = Element.BOARD+'HexGrid'
    def __init__(self,zone,node=None,**kwargs):
        super(HexGrid,self).__init__(zone,self.TAG,node=node,**kwargs)

    def addNumbering(self,**kwargs):
        '''Add a `Numbering` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Numbering
            The added element
        '''
        return self.add(HexNumbering,**kwargs)
    def getNumbering(self):
        return self.getAllElements(HexNumbering)
    def getDeltaX(self):
        return float(self['dx'])
    def getDeltaY(self):
        return float(self['dy'])
    def getXOffset(self):
        return int(self['x0'])
    def getYOffset(self):
        return int(self['y0'])
    def getMaxRows(self):
        from math import floor
        height    = self.getZone().getHeight()
        return floor(height / self.getDeltaX() + .5)
    def getMaxCols(self):
        from math import floor
        width    = self.getZone().getWidth()
        return floor(width / self.getDeltaY()  + .5)

registerElement(HexGrid)

# --------------------------------------------------------------------
class SquareGrid(BaseGrid):
    TAG = Element.BOARD+'SquareGrid'
    def __init__(self,zone,node=None,
                 dx           = RECT_WIDTH,
                 dy           = RECT_HEIGHT,
                 edgesLegal   = False,
                 x0           = 0,
                 y0           = int(0.4*RECT_HEIGHT),
                 **kwargs):
        super(SquareGrid,self).__init__(zone,self.TAG,node=node,
                                        dx         = dx,
                                        dy         = dy,
                                        edgesLegal = edgesLegal,
                                        x0         = x0,
                                        y0         = y0,
                                        **kwargs)
    def addNumbering(self,**kwargs):
        '''Add a `Numbering` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Numbering
            The added element
        '''
        return self.add(SquareNumbering,**kwargs)
    def getNumbering(self):
        return self.getAllElements(SquareNumbering)
    def getDeltaX(self):
        return float(self['dx'])
    def getDeltaY(self):
        return float(self['dy'])
    def getXOffset(self):
        return int(self['x0'])
    def getYOffset(self):
        return int(self['y0'])
    def getMaxRows(self):
        from math import floor
        height    = self.getZone().getHeight()
        return floor(height / self.getDeltaY() + .5)
    def getMaxCols(self):
        from math import floor
        width    = self.getZone().getWidth()
        return floor(width / self.getDeltaX()  + .5)

registerElement(SquareGrid)

# --------------------------------------------------------------------
class HexNumbering(BaseNumbering):
    TAG = Element.BOARD+'mapgrid.HexGridNumbering'
    def __init__(self,grid,node=None,**kwargs):
        super(HexNumbering,self).__init__(grid,self.TAG,node=node,**kwargs)
        
    def getGrid(self):
        g = self.getParent(HexGrid)
        return g

    def _getCenter(self,col,row):
        '''Convert col and row index to picture coordinates'''
        from math import floor
        
        stagger  = self['stagger'] == 'true'
        sideways = self.getGrid()['sideways'] == 'true'
        hDesc    = self['hDescend'] == 'true'
        vDesc    = self['vDescend'] == 'true'
        xOff     = self.getGrid().getXOffset()
        yOff     = self.getGrid().getYOffset()
        hexW     = self.getGrid().getDeltaX()
        hexH     = self.getGrid().getDeltaY()
        zxOff    = self.getGrid().getZone().getXOffset()
        zyOff    = self.getGrid().getZone().getYOffset()
        maxRows  = self.getGrid().getMaxRows()
        maxCols  = self.getGrid().getMaxCols()
        # print(f'  Col:         {col}')
        # print(f'  Row:         {row}')
        # print(f'  Stagger:     {stagger}')
        # print(f'  Sideways:    {sideways}')
        # print(f'  hDesc:       {hDesc}')
        # print(f'  vDesc:       {vDesc}')
        # print(f'  maxRows:     {maxRows}')
        # print(f'  maxCols:     {maxCols}')

        # This code from HexGridNumbering.java
        if stagger:
            if sideways:
                if col % 2 != 0:
                    if hDesc:
                        row += 1
                    else:
                        row -= 1
            else:
                if col % 2 != 0:
                    if vDesc:
                        row += 1
                    else:
                        row -= 1

        if sideways:
            if vDesc:
                col = maxRows - col
            if hDesc:
                row = maxCols - row
        else:
            if hDesc:
                col = maxCols - col
            if vDesc:
                row = maxRows - row


        x = col * hexW + xOff
        if col % 2 == 0:
            y = row * hexH
        else:
            y = row * hexH + hexH / 2
        y += yOff

        # print(f'  Col:         {col}')
        # print(f'  Row:         {row}')
        # print(f'  hexW:        {hexW}')
        # print(f'  hexH:        {hexH}')
        # print(f'  xOff:        {xOff}')
        # print(f'  yOff:        {yOff}')
        # print(f'  x:           {x}')
        # print(f'  y:           {y}')
        
        if sideways:
            x, y = y, x

        return int(floor(x+.5)),int(floor(y+.5))
            
        # if sideways:
        #     maxRows, maxCols = maxCols, maxRows
        #     
        # if stagger:
        #     if sideways:
        #         if col % 2 != 0:
        #             row += 1 if hDesc else -1
        #     else:
        #         if col % 2 != 0:
        #             row += 1 if vDesc else -1
        # 
        # if hDesc:
        #     col = maxCols - col
        # if vDesc:
        #     row = maxRows - row
        # 
        # print(f'  Col:         {col}')
        # print(f'  Row:         {row}')
        # print(f'  hexW:        {hexW}')
        # print(f'  hexH:        {hexH}')
        # print(f'  xOff:        {xOff}')
        # print(f'  yOff:        {yOff}')
        # 
        # x = col * hexW + xOff
        # y = row * hexH + yOff + (hexH/2 if col % 2 != 0 else 0)
        # 
        # print(f'  Col:         {col}')
        # print(f'  Row:         {row}')
        # print(f'  hexW:        {hexW}')
        # print(f'  hexH:        {hexH}')
        # print(f'  xOff:        {xOff}')
        # print(f'  yOff:        {yOff}')
        # print(f'  x:           {x}')
        # print(f'  y:           {y}')
        # 
        # x = int(floor(x + .5))
        # y = int(floor(y + .5))
        # if sideways:
        #     # print(f'Swap coordinates because {sideways}')
        #     x, y = y, x
        # 
        # return x,y

registerElement(HexNumbering)

# --------------------------------------------------------------------
class SquareNumbering(BaseNumbering):
    TAG = Element.BOARD+'mapgrid.SquareGridNumbering'
    def __init__(self,grid,node=None,hType='N',**kwargs):
        super(SquareNumbering,self).__init__(grid,self.TAG,node=node,
                                             hType=hType,**kwargs)
    def getGrid(self):
        return self.getParent(SquareGrid)

    def getCenter(self,col,row):
        hDesc    = self['hDescend'] == 'true'
        vDesc    = self['vDescend'] == 'true'
        xOff     = self.getGrid().getXOffset()
        yOff     = self.getGrid().getYOffset()
        squareW  = self.getGrid().getDeltaX()
        squareH  = self.getGrid().getDeltaY()
        maxRows  = self.getGrid().getMaxRows()
        maxCols  = self.getGrid().getMaxCols()

        if vDesc:  row = maxRows - row
        if hDesc:  col = maxCols - col

        x = col * squareW + xOff
        y = row * squareH + yOff

        return x,y
        
registerElement(SquareNumbering)
    
# --------------------------------------------------------------------
class RegionGrid(Element):
    TAG = Element.BOARD+'RegionGrid'
    def __init__(self,zone,node=None,snapto=True,fontsize=9,visible=True):
        super(RegionGrid,self).__init__(zone,self.TAG,node=node,
                                        snapto   = snapto,
                                        fontsize = fontsize,
                                        visible  = visible)

    def getZone(self):
        ## BEGIN_IMPORT
        from . zone import Zone
        ## END_IMPORT
        return self.getParent(Zone)
    def getZoneGrid(self):
        z = self.getZone()
        if z is not None:
            return z.getBoard()
        return None
    def getBoard(self):
        z = self.getZonedGrid()
        if z is not None:
            return z.getBoard()
        return self.getParent(Board)
    def getMap(self):
        b = self.getBoard()
        if b is not None:
            return b.getMap()
        return None
    def getRegions(self):
        return self.getElementsByKey(Region,'name')    
    def checkName(self,name):
        '''Get unique name'''
        poss = len([e for e in self.getRegions()
                    if e == name or e.startswith(name+'_')])
        if poss == 0:
            return name

        return name + f'_{poss}'
    def addRegion(self,**kwargs):
        '''Add a `Region` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Region
            The added element
        '''
        return self.add(Region,**kwargs)
    def getLocation(self,loc):
        for r in self.getRegions().values():
            if loc == r['name']:
                return int(r['originx']),int(r['originy'])

        return None
        
registerElement(RegionGrid)

# --------------------------------------------------------------------
class Region(Element):
    TAG = Element.BOARD+'Region'
    UNIQUE = ['name']
    def __init__(self,grid,node=None,
                 name      = '',
                 originx   = 0,
                 originy   = 0,
                 alsoPiece = True,
                 piece     = None,
                 prefix    = ''):
        fullName = name + ("@"+prefix if len(prefix) else "")
        realName = grid.checkName(fullName) if node is None else fullName
        super(Region,self).__init__(grid,
                                    self.TAG,
                                    node    = node,
                                    name    = realName,
                                    originx = originx,
                                    originy = originy)

        if node is None and alsoPiece:
            m = self.getMap()
            b = self.getBoard()
            if m is not None and b is not None:
                if piece is None:
                    g      = m.getGame()
                    pieces = g.getSpecificPieces(name,asdict=False)
                    piece  = pieces[0] if len(pieces) > 0 else None
             
                if piece is not None:
                    # bname = m['mapName']
                    bname = b['name']
                    #print(f'Adding at-start name={name} location={realName} '
                    #      f'owning board={bname}')
                    a = m.addAtStart(name            = name,
                                     location        = realName,
                                     useGridLocation = True,
                                     owningBoard     = bname,
                                     x               = 0,
                                     y               = 0)
                    p = a.addPiece(piece)
                    if p is None:
                        print(f'EEE Failed to add piece {name} ({piece}) to add-start {a}')
                    #if p is not None:
                    #    print(f'Added piece {name} in region')
                #else:
                #    print(f'Could not find piece {name}')
            
    def getGrid(self):
        return self.getParent(RegionGrid)
    def getZone(self):
        g = self.getGrid()
        if g is not None:
            return g.getZone()
        return None
    def getZonedGrid(self):
        z = self.getZone()
        if z is not None:
            return z.getZonedGrid()
        return None
    def getBoard(self):
        z = self.getZonedGrid()
        if z is not None:
            return z.getBoard()
        return self.getParent(Board)
    def getPicker(self):
        z = self.getBoard()
        if z is not None:
            return z.getPicker()
        return None
    def getMap(self):
        b = self.getPicker()
        if b is not None:
            return b.getMap()
        return None

registerElement(Region)

#
# EOF
#



#!/usr/bin/env python
import sys
sys.path.append('..')

from common import Verbose, VerboseGuard
from common.collector import Collector

        
if __name__ == '__main__':
    from argparse import ArgumentParser, FileType  

    ap = ArgumentParser(description='Collect to single script')
    ap.add_argument('output',
                    type=FileType('w'),
                    nargs='?',
                    default='gbx0pwd.py',
                    help='Output script name')
    ap.add_argument('-v','--verbose',action='store_true',
                    help='Be verbose')

    args = ap.parse_args()

    Verbose().setVerbose(args.verbose)

    c = Collector()
    c.run(args.output,
          'cyberboard.py',
          'zeropwd.py')

    



## BEGIN_IMPORT
from . verboseguard import VerboseGuard
from . verbose import Verbose
## END_IMPORT


class Collector:
    def __init__(self,executable=True):
        '''This could be done with SED, but we don't need to rely on that'''
        self._executable = executable

    def readFile(self,filename):
        '''Reads lines from a file and returns them.

        Lines fenced by with `##` are not returned.

        Lines starting with `#!` are not returned
        '''
        with VerboseGuard(f'Reading lines from {filename}') as g:
            ret = [f'# {"="*68}\n'
                   f'# From {filename}\n' ]
            with open(filename,'r') as file:
                lines  = file.readlines()
                nlines = len(lines)
                lineno = 0
            
                g(f'{filename} has {nlines} lines')
                while lineno < nlines:
                    line = lines[lineno]
                    if line.strip().startswith('##'):
                        start   = line
                        startno = lineno
            
                        while lineno < nlines:
                            lineno += 1
                            if lineno >= nlines:
                                raise RuntimeError(f'At {filename}:{lineno}:'
                                                   f' No end to {start} in '
                                                   f'line {startno}')
                            
                            line   =  lines[lineno]
                            if line.strip().startswith('##'):
                                g(f'Skipped lines {startno} to {lineno}')
                                break
                        lineno += 1
                        continue
            
                    if line.startswith('#!'):
                        g(f'Skipping line {lineno}')
                        lineno += 1
                        continue
            
                    ret.append(line)
                    lineno += 1
            
                return ret

    def run(self,output,*filenames):
        lines = ['#!/usr/bin/env python\n',
                 '# Script collected from other scripts\n',
                 '#\n']
        lines.extend([f'#   {filename}\n' for filename in filenames])
        lines.append('#\n')
            

        for filename in filenames:
            lines.extend(self.readFile(filename))

        lines.extend(['##\n',
                      '# End of generated script\n'
                      '##\n'])


        output.writelines(lines)

        if not self._executable:
            return
        
        from os import chmod
        chmod(output.name,0o755)


#
# EOF
#

            



#!/usr/bin/env python
#
# CyberBoard GitHub page:
#
#    https://github.com/CyberBoardPBEM/cbwindows
#
#
## BEGIN_IMPORT
from pywargame.cyberboard.gamebox   import GameBox
from pywargame.cyberboard.extractor import GBXExtractor
from pywargame.common               import Verbose
## END_IMPORT

# ====================================================================
def extractMain():
    from argparse import ArgumentParser, FileType
    from pathlib import Path

    ap = ArgumentParser(description='Extract information from a '
                        'CyberBoard GameBox file and store in ZIP archive',
                        epilog='Default output file name is input file name '
                        'with .gbx replaced by .zip')
    ap.add_argument('input', type=str, help='The file')
    ap.add_argument('output',type=str, nargs='?',help='Output',default='')
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ap.add_argument('-D','--dump',type=str,nargs='*',default='',
                    help='Dump content')
    ap.add_argument('-S','--size-bytes',type=int,choices=[4,8],
                    default=4, help='The size of size counts')
    ap.add_argument('-I','--id-bytes',type=int,choices=[2,4],
                    default=2, help='The size of identifiers')
    ap.add_argument('-P','--password',action='store_true',
                    help='Show password')

    args = ap.parse_args()
    Verbose().setVerbose(args.verbose)

    Features().size_size = args.size_bytes
    Features().id_size   = args.id_bytes
    
    gbx = GameBox.fromFile(args.input)
    if args.password:
        print(f'Password is "{gbx._info._password.decode()}"')
        
    rat = GBXExtractor(gbx)

    out = args.output
    if out == '':
        p = Path(args.input)
        out = p.stem+'.zip'
    
    rat.save(out)

    if 'save' in args.dump:
        print(rat)
    if 'gbx' in args.dump:
        print(gbx)
    

# ====================================================================
if __name__ == '__main__':
    extractMain()

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . draw import GBXDrawList
from . cell import GBXCell, GBXCellGeometry
from . base import CbManager, CbFont
from . head import num_version
from . features import Features
from . draw import GBXDraw
## END_IMPORT

# --------------------------------------------------------------------
class GBXBoardCalculator:
    def __init__(self,board):
        self._geometry  = board._full
        self._nRows     = board._nRows
        self._nCols     = board._nCols
        self._rowOffset = board._rowOffset
        self._colOffset = board._colOffset
        self._rowInvert = board._rowInvert
        self._colInvert = board._colInvert

    def __call__(self,x,y):
        # Shift depending on grid type and stagger
        row, col = self._geometry.inverse(x,y)
        if self._rowInvert:
            row = self._nRows - row - 1
        if self._colInvert:
            col = self._nCols - col - 1

        return row+self._rowOffset, col+self._colOffset

# --------------------------------------------------------------------
class GBXBoard:
    def __init__(self,ar):
        '''A board'''
        with VerboseGuard(f'Reading board') as g:
            self._serial          = ar.iden()
            self._visible         = ar.word()
            self._snap            = ar.word()
            self._xSnap           = ar.dword()
            self._ySnap           = ar.dword()
            self._xSnapOffset     = ar.dword()
            self._ySnapOffset     = ar.dword()
            self._maxLayer        = ar.word()
            self._background      = ar.dword()
            self._name            = ar.str()
            hasDraw               = ar.word()
            self._baseDraw        = GBXDrawList(ar) if hasDraw else None
                
            self._showCellBorder  = ar.word()
            self._topCellBorder   = ar.word()
            self._reserved        = [ar.word() for _ in range(4)]
            self._reserved2       = None
            self._rowOffset       = 0
            self._colOffset       = 0
            self._rowInvert       = False
            self._colInvert       = False
            self._nRows           = 0
            self._nCols           = 0
            self._transparent     = False
            self._numbers         = 0
            self._trackCell       = False
            self._frameColor      = 0xFF000000
            self._full            = None
            self._half            = None
            self._small           = None
            self._map             = []
            self._topDraw         = None
            
            hasArray              = ar.word()
            if hasArray != 0:
                self._reserved2   = [ar.word() for _ in range(4)]
                self._rowOffset   = ar.word()
                self._colOffset   = ar.word()
                self._rowInvert   = ar.word()
                self._colInvert   = ar.word()
                self._nRows       = ar.int(Features().sub_size)
                self._nCols       = ar.int(Features().sub_size)
                self._transparent = ar.word()
                self._numbers     = ar.word()
                self._trackCell   = ar.word()
                self._frameColor  = ar.dword()
            
                self._full        = GBXCellGeometry(ar)
                self._half        = GBXCellGeometry(ar)
                self._small       = GBXCellGeometry(ar)
            
                self._map = [[GBXCell(ar,row,col) for col in range(self._nCols)]
                             for row in range(self._nRows)]
            
            hasDraw              = ar.word()
            self._topDraw        = GBXDrawList(ar) if hasDraw else None
            g(f'Board background read: {self._background:06x}, frame color: {self._frameColor:06x}')

    def toDict(self,tileManager,markManager,strings,no,boardDigits,
               alsoMap=True):
        from io import StringIO
        
        with VerboseGuard(f'Making dict of board {self._name}') as g:
            sav = f'board_{no:0{boardDigits}d}.svg'
            g(f'File to save in: {sav}')
            dct = {'name':               self._name,
                   'serial':             self._serial,
                   'visible':            self._visible,
                   'snap':  {
                       'enable':         self._snap,
                       'x': { 
                           'distance':   self._xSnap,
                           'offset':     self._xSnapOffset
                       },
                       'y': {
                           'distance':   self._ySnap,
                           'offset':     self._ySnapOffset
                       }
                   },
                   'max layer':          self._maxLayer,
                   'cell border': {
                       'visible':        self._showCellBorder,
                       'on top layer':   self._topCellBorder,
                   },
                   'rows': {
                       'size':           self._nRows,
                       'offset':         self._rowOffset,
                       'inverted':       self._rowInvert
                   },
                   'columns': {
                       'size':           self._nCols,
                       'offset':         self._colOffset,
                       'inverted':       self._colInvert
                   },
                   'cells': {
                       'transparent':    self._transparent,
                       'foreground':     self._frameColor,
                   },
                   'numbering': {
                       'order':   'V' if self._numbers % 2 == 0 else 'H',
                       'padding': self._numbers in [2,3],
                       'first':   'A' if self._numbers in [4,5] else 'N'
                   }
                }
            if self._full is not None:
                dct['cells']['geometry'] = self._full.toDict()
            if alsoMap and self._map is not None:
                dct['cells']['list'] = [[c.toDict(tileManager,self._full)
                                         for c in row]
                                        for row in self._map]
            
            
            sav = f'board_{no:0{boardDigits}d}.svg'
            img = self.drawing(sav,
                               tileManager=tileManager,
                               markManager=markManager)
            # img.save(pretty=True)

            stream = StringIO()
            img.write(stream,pretty=True)
            
            dct['filename'] = sav
            dct['image']    = stream.getvalue()#img.tostring()
            dct['size']     = self._full.boardSize(self._nRows,self._nCols)
            
            return dct

    def drawing(self,sav,tileManager,markManager,*args,**kwargs):
        from svgwrite import Drawing
        from svgwrite.base import Title
        with VerboseGuard(f'Making SVG of board {self._name}') as g:
            size  = self._full.boardSize(self._nRows,self._nCols)
            dwg   = Drawing(filename=sav,size=size)
            frame, defMap, patMap = self.defs(dwg,tileManager,markManager)
            
            # Draw background
            g(f'Board background: {self._background:06x} {GBXDraw.hex(self._background)}')
            dwg.add(Title(self._name))
            dwg.add(dwg.rect(id='background',
                             insert=(0,0),
                             size=size,
                             fill=GBXDraw.hex(self._background)
                             #f'#{self._background:06x}')
                             ))
            # GBXDraw.hex(self._background)

            g('Drawing base layer')
            bse  = self.base (dwg, 0, defMap)
            g('Drawing cells')
            grd  = self.cells(dwg, frame, patMap)
            if self._showCellBorder and not self._topCellBorder:
                self.borders(dwg,frame)
            g('Drawing top layer')
            top1 = self.top  (dwg, 1, defMap)
            if self._showCellBorder and self._topCellBorder:
                self.borders(dwg,frame)
            top2 = self.top  (dwg, 2, defMap)
                    
            return dwg

    def defs(self,dwg,tileManager,markManager):
        defMap         = {}
        patMap         = {}
        frame          = self._full.svgDef(dwg)
        defMap['cell'] = frame

        # Get defininitions from base layer 
        if self._baseDraw:
            self._baseDraw.svgDefs(dwg,
                                   tileManager,
                                   markManager,
                                   defMap)
        # Get definitions from cell layer 
        for row in self._map:
            for cell in row:
                cell.svgDef(dwg,tileManager,patMap)
            
        # Get definitions from top layer 
        if self._topDraw:
            self._topDraw.svgDefs(dwg,
                                  tileManager,
                                  markManager,
                                  defMap)

        return frame, defMap, patMap

    def base(self,dwg,passNo,defMap):
        bse  = dwg.add(dwg.g(id=f'base_{passNo:02d}'))
        if self._baseDraw:
            self._baseDraw.svg(dwg,bse,passNo,defMap)

        return bse
        
    def cells(self,dwg,frame,patMap):
        grd  = dwg.add(dwg.g(id='grid'))
        for row in self._map:
            for cell in row:
                cell.svg(dwg,grd,frame,self._full,patMap)

        return grd

    def borders(self,dwg,frame):
        brd  = dwg.add(dwg.g(id='borders'))
        for row in self._map:
            for cell in row:
                cell.svgFrame(dwg,brd,frame,self._full,self._frameColor)

        return brd
        
    def top(self,dwg,passNo,defMap):
        top  = dwg.add(dwg.g(id=f'top_{passNo:02d}'))
        if self._topDraw:
            self._topDraw.svg (dwg,top,passNo,defMap)
        
        return top

    def __str__(self):
        return (f'GBXBoard: {self._name}\n'
                f'      serial:           {self._serial}\n'
                f'      visible:          {self._visible}\n'
                f'      snap:             {self._snap}\n'
                f'      xSnap:            {self._xSnap}\n'
                f'      ySnap:            {self._ySnap}\n'
                f'      xSnapOffset:      {self._xSnapOffset}\n'
                f'      ySnapOffset:      {self._ySnapOffset}\n'
                f'      maxLayer:         {self._maxLayer}\n'
                f'      background:       {self._background:08x}\n'
                f'      Base draws:       {self._baseDraw}\n'
                f'      Show cell border: {self._showCellBorder}\n'
                f'      Top cell border:  {self._topCellBorder}\n'
                f'      Reserved:         {self._reserved}\n'
                f'      Reserved2:        {self._reserved}\n'
                f'      Row offset:       {self._rowOffset}\n'
                f'      Column offset:    {self._colOffset}\n'
                f'      Row invert:       {self._rowInvert}\n'
                f'      Colunn invert:    {self._colInvert}\n'
                f'      # Rows:           {self._nRows}\n'
                f'      # Cols:           {self._nCols}\n'
                f'      Transparent:      {self._transparent}\n'
                f'      Numbers:          {self._numbers}\n'
                f'      Track cells:      {self._trackCell}\n'
                f'      Frame color:      {self._frameColor:08x}\n'
                f'      Full geometry:    {self._full}\n'
                f'      Half geometry:    {self._half}\n'
                f'      Small geometry:   {self._small}\n'
                f'      Top draws:        {self._topDraw}'
                )
    
            
            
        
# --------------------------------------------------------------------
class GBXBoardManager(CbManager):
    def __init__(self,ar):
        '''Manager of boards'''
        with VerboseGuard(f'Reading board manager'):
            self._nextSerial  = ar.iden()
            super(GBXBoardManager,self).__init__(ar)
            # print(Features().id_size)
            self._boards = self._readSub(ar,GBXBoard)

    def __len__(self):
        return len(self._boards)

    def bySerial(self,serial):
        for b in self._boards:
            if b._serial == serial:
                return b

        return None
    
    def toDict(self,tileManager,markManager,strings):
        from math import log10, ceil
        with VerboseGuard(f'Making dict board manager'):
            boardDigits = int(ceil(log10(len(self)+.5)))

            return {b._serial:
                    b.toDict(tileManager,markManager,strings,no,boardDigits)
                    for no, b in enumerate(self._boards)}
    
    def __str__(self):
        return ('GBXBoard manager:\n'
                + f'  Next serial:  {self._nextSerial:08x}\n'
                + super(GBXBoardManager,self).__str__()
                + self._strSub('boards',self._boards))

# --------------------------------------------------------------------
class GSNGeomorphicElement:
    def __init__(self,ar):
        with VerboseGuard('Reading geomorphic element'):
            self._serial = ar.word()
        
    def __str__(self):
        return f'GSNGeomorphicElement: {self._serial}'
    
# --------------------------------------------------------------------
class GSNGeomorphicBoard:
    def __init__(self,ar):
        with VerboseGuard('Reading geomorphic board'):
            self._name     = ar.str()
            self._nRows    = ar.word()
            self._nCols    = ar.word()
            n              = ar.word()
            self._elements = [GSNGeomorphicElement(ar) for _ in range(n)]
        
    def __str__(self):
        pl = '\n    '.join([str(s) for s in self._elements])
        return (f'GeomorphicBoard: {self._name}\n'
                f'  Size: {self._nRows}x{self._nCols}\n'
                f'  Elements:\n    {pl}\n')
    
# --------------------------------------------------------------------
class GSNBoard:
    def __init__(self,ar,vers):
        with VerboseGuard(f'Reading scenario board {vers//256}.{vers%256}'):
            hasGeo                 = ar.byte()
            self._geo              = GSNGeomorphicBoard(ar) if hasGeo else None
            self._serial           = ar.iden()
            self._snap             = ar.word()
            self._xSnap            = ar.dword()
            self._ySnap            = ar.dword()
            self._xSnapOffset      = ar.dword()
            self._ySnapOffset      = ar.dword()
            self._xStagger         = ar.word()
            self._yStagger         = ar.word()
            self._piecesVisible    = ar.word()
            self._blockBeneath     = ar.word()
            self._rotate180        = ar.word()
            self._showTiny         = ar.word()
            self._indicatorsVisible= ar.word()
            self._cellBorders      = ar.word()
            self._smallCellBorders = ar.word()
            self._enforceLocks     = ar.word()
            self._plotLineColor    = ar.dword()
            self._plotLineWidth    = ar.word()
            self._lineColor        = ar.dword()
            self._lineWidth        = ar.word()
            self._textColor        = ar.dword()
            self._textBoxColor     = ar.dword()
            self._font             = CbFont(ar)
            self._gridCenters      = ar.word()
            self._snapMove         = ar.word()
            self._indactorsTop     = ar.word()
            self._openOnLoad       = ar.word()
            self._prevPlotMode     = ar.word()
            self._prevPlotX        = ar.word()
            self._prevPlotY        = ar.word()
            self._ownerMask        = (ar.word() if vers < num_version(3,10) else
                                      ar.dword())
            self._restrictToOwner  = ar.word()
            self._pieces           = GBXDrawList(ar)
            self._indicators       = GBXDrawList(ar)
        
        
    def toDict(self,boardManager):
        board = (None if boardManager is None else
                 boardManager.bySerial(self._serial))
        geom  = None if board is None else board._full
        calc  = None if geom  is None else GBXBoardCalculator(board)
        
        return {
            'onload':             self._openOnLoad != 0,
            'snap':  {
                'enable':         self._snap,
                'onmove':         self._snapMove != 0,
                'gridCenter':     self._gridCenters != 0,
                'x': { 
                    'distance':   self._xSnap,
                    'offset':     self._xSnapOffset
                },
                'y': {
                    'distance':   self._ySnap,
                    'offset':     self._ySnapOffset
                }
            },
            'moves': {
                'color':          self._plotLineColor,
                'width':          self._plotLineWidth
            },
            'stacking':           [self._xStagger, self._yStagger],
            'owner':              self._ownerMask,
            'restrict':           self._restrictToOwner != 0,
            'grid': {
                'show':           self._cellBorders != 0,
                'color':          self._lineColor,
                'width':          self._lineWidth
            },
            'pieces':             self._pieces.toDict(calc),
            'indicators':         self._indicators.toDict(calc)
        }
    
    def __str__(self):
        return (f'ScenarioBoard: {self._serial}'
                f'      Geomorphic: {"None" if self._geo is None else self._geo}\n'
                f'      Font:       {self._font}\n'
                f'      Pieces:\n{str(self._pieces)}\n'
                f'      Indicators:\n{str(self._indicators)}')
    
# --------------------------------------------------------------------
class GSNBoardManager:    
    def __init__(self,ar,vers):
        with VerboseGuard(f'Reading scenario board manager') as g:
            self._nextGeoSerial = ar.iden()
            self._reserved      = [ar.word() for _ in range(3)]
            n                   = ar.sub_size()
            g(f'Got {n} boards to read')
            self._boards        = [GSNBoard(ar,vers) for _ in range(n)]

    def toDict(self,boardManager):
        hasStart = False
        for b in self._boards:
            if b._openOnLoad:
                hasStart = True

        # Make sure at least one map is loaded 
        if not hasStart and len(self._boards) > 0:
            self._boards[0]._openOnLoad = True
            
        return {b._serial: b.toDict(boardManager) for b in self._boards }

    def __str__(self):
        pl = '\n    '.join([str(s) for s in self._boards])
        return f'GSNBoardManager: {self._nextGeoSerial}\n    {pl}\n'
    
#
# EOF
#




## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class ClickTrait(Trait):
    ID = 'button'
    def __init__(self,
                 key         = '',
                 x           = 0,
                 y           = 0,
                 width       = 0,
                 height      = 0,
                 description = '',
                 context     = False,
                 whole       = True,
                 version     = 1,
                 points      = []):
        '''Create a click trait (static property)'''
        super(ClickTrait,self).__init__()
        self.setType(key          = key,
                     x            = x,
                     y            = y,
                     width        = width,
                     height       = height,
                     description  = description,
                     context      = context,
                     whole        = whole,
                     version      = version,
                     npoints      = len(points),
                     points       = ';'.join([f'{p[0]};{p[1]}'
                                              for p in points]))            
        self.setState()


Trait.known_traits.append(ClickTrait)

#
# EOF
#



#!/usr/bin/env python
#
# CyberBoard GitHub page:
#
#    https://github.com/CyberBoardPBEM/cbwindows
#
            
# ====================================================================
if __name__ == '__main__':
    from argparse import ArgumentParser, FileType
    from pathlib import Path
## BEGIN_IMPORT
    from gamebox   import GameBox
    from extractor import Extractor
    from common    import Verbose
## END_IMPORT

    ap = ArgumentParser(description='Read the file')
    ap.add_argument('input', type=str, help='The file')
    ap.add_argument('output',type=str, nargs='?',help='Output',default='')
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ap.add_argument('-D','--dump',type=str,nargs='*',default='',
                    help='Dump content')

    args = ap.parse_args()
    Verbose().setVerbose(args.verbose)
    
    gbx = GameBox.fromFile(args.input)
    rat = Extractor(gbx)

    out = args.output
    if out == '':
        p = Path(args.input)
        out = p.stem+'.zip'
    
    rat.save(out)

    if 'save' in args.dump:
        print(rat)
    if 'gbx' in args.dump:
        print(gbx)

    

    



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
## END_IMPORT

# --------------------------------------------------------------------
class Command:
    def __init__(self,what,iden,tpe,state):
        self.cmd = '/'.join([what,iden,tpe,state])
    
# --------------------------------------------------------------------
class AddCommand(Command):
    ID = '+'
    def __init__(self,iden,tpe,state):
        super(AddCommand,self).__init__(self.ID,iden,tpe,state)
        

#
# EOF
#



#!/usr/bin/env python

# --------------------------------------------------------------------
def diceMain():
    from sys import path
    
    from argparse import ArgumentParser
    from pywargame.common.dicedraw import DiceDrawer

    ap = ArgumentParser(description='Make a series of dice images')
    ap.add_argument('-n','--sides', choices=[4,6,8,10,12,20],
                    default=6, type=int,
                    help='Number of sides')
    ap.add_argument('-f','--foreground', type=str, default='white',
                    help='Foreground color')
    ap.add_argument('-b','--background', type=str, default='#333333',
                    help='Background color')
    ap.add_argument('-W','--width',type=int,default=75,
                    help='Width of images')
    ap.add_argument('-H','--height',type=int,default=75,
                    help='Height of images')
    ap.add_argument('-B','--base',type=str,default='d{sides}-{value}.png',
                    help='Format of file names')

    args = ap.parse_args()

    dd = DiceDrawer(args.sides,
                    args.width,
                    args.height,
                    fg = args.foreground,
                    bg = args.background)

    base = args.base

    vals = list(range(0,args.sides) if args.sides == 10 else
                range(1,args.sides+1))
    for val in vals:
        dd.draw(val).save(filename=base.format(sides=args.sides,value=val))
    
# --------------------------------------------------------------------
if __name__ == '__main__':
    diceMain()

# --------------------------------------------------------------------
#
# EOF
#

    



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class HideTrait(Trait):
    ID           = 'hide'
    ANY_SIDE     = 'side:'
    ANY_PLAYER   = 'player:'
    SIDES        = 'sides:'
    @classmethod
    def encodeAccess(cls,spec):
        if isinstance(spec,list) and len(spec) == 1:
            spec = spec[0]
        if isinstance(spec,str):
            if spec == cls.ANY_SIDE: return cls.ANY_SIDE
            if spec == cls.ANY_PLAYER: return cls.ANY_PLAYER
            return cls.SIDES+":"+spec
        return cls.SIDES+':'.join(spec)

    @classmethod
    def decodeAccess(cls,spec):
        if spec.startswith(cls.ANY_SIDE):   return cls.ANY_SIDE
        if spec.startswith(cls.ANY_PLAYER): return cls.ANY_PLAYER
        if spec.startswith(cls.SIDES):      return spec.split(':')[1:]
        return None
        
    def __init__(self,
                 key                   = '',
                 command               = '',
                 bgColor               = rgb(0x0,0x0,0x0),
                 access                = [],
                 transparency          = 1, # between 0 and 1
                 description           = '',
                 disableAutoReportMove = False,
                 state                 = 'null'):
        '''Create a hide trait (VASSAL.counter.Hideable)'''

        super().__init__()
        spec = self.encodeAccess(access)
        
        self.setType(key                   = key,      # Context menu key
                     command               = command,  # Context menu name
                     bgColor               = bgColor,
                     access                = spec,
                     transparency          = transparency,
                     description           = description,
                     disableAutoReportMove = disableAutoReportMove)     
        self.setState(hiddenBy = state)

    def getAccess(self):
        return self.decodeAccess(self['access'])

    def setAccess(self, access = []):
        self['access'] = self.encodeAccess(access)

        

        

Trait.known_traits.append(HideTrait)

#
# EOF
#



# --------------------------------------------------------------------
## BEGIN_IMPORT
from .. common.singleton import Singleton
## END_IMPORT

class Features(metaclass=Singleton):
    def __init__(self):
        self.bmp_zlib      = False # wxBMPHandler + Zlib
        self.id_size       = 2     # Size of IDs in bytes (1, 2 or 4)
        self.size_size     = 4     # Size of sizes in bytes (4 or 8)
        self.sub_size      = 2     # Size of sub-sizes in bytes (4 or 8)
        self.square_cells  = False # Geomorphic boards, square cells
        self.rotate_unit   = False # Geomorphic boards, rotated unit board
        self.piece_100     = False # Pieces w/<= 100 sides
        self.private_board = False #
        self.roll_state    = False # serialize roll state
        self.little_endian = True



'''
Module to manipulate a VASSAL module (VMOD)
'''
from . base import *
from . element import *
from . folder import *
from . globalkey import *
from . gameelements import *
from . mapelements import *
from . globalproperty import *
from . turn import *
from . documentation import *
from . player import *
from . widget import *
from . grid import *
from . zone import *
from . board import *
from . map import *
from . chart import *
from . command import *
from . trait import *
from . withtraits import *
from . extension import *
from . traits.area import *
from . traits.clone import *
from . traits.dynamicproperty import *
from . traits.globalproperty import *
from . traits.prototype import *
from . traits.place import *
from . traits.report import *
from . traits.calculatedproperty import *
from . traits.restrictcommand import *
from . traits.label import *
from . traits.layer import *
from . traits.globalcommand import *
from . traits.globalhotkey import *
from . traits.nostack import *
from . traits.deselect import *
from . traits.restrictaccess import *
from . traits.rotate import *
from . traits.stack import *
from . traits.mark import *
from . traits.mask import *
from . traits.trail import *
from . traits.delete import *
from . traits.sendto import *
from . traits.moved import *
from . traits.skel import *
from . traits.submenu import *
from . traits.basic import *
from . traits.trigger import *
from . traits.nonrect import *
from . traits.click import *
from . traits.mat import MatTrait
from . traits.cargo import CargoTrait
from . traits.movefixed import MoveFixedTrait
from . traits.sheet import SheetTrait
from . traits.hide import HideTrait
from . traits.retrn import ReturnTrait
from . game import *
from . buildfile import *
from . moduledata import *
from . save import *
from . vsav import *
from . vmod import *
from . exporter import *
from . merger import Merger



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . game import Game
## END_IMPORT

# ====================================================================
class Extension(Element):
    TAG = Element.MODULE+'ModuleExtension'
    def __init__(self,
                 parent          = None,
                 node            = None,
                 anyModule       = False,
                 version         = '',
                 description     = '',
                 module          = '',
                 moduleVersion   = '',
                 vassalVersion   = '',
                 nextPieceSlotId = 0,
                 extensionId     = 0,
                 asDocument      = False):
        super().__init__(parent,self.TAG,node)

        self._tag  = 'extension'
        if self._node is None:
            #from xml.dom.minidom import Document


            self._root = xmlns.Document()
            self._node = self._root
            self.setAttributes(
                anyModule       = anyModule,
                version         = version,
                description     = description,
                module          = module,
                moduleVersion   = moduleVersion,
                vassalVersion   = vassalVersion,
                nextPieceSlotId = nextPieceSlotId,
                extensionId     = extensionId)

    def addExtensionElement(self,**kwargs):
        '''Add an extension element'''

        return self.add(ExtensionElement,**kwargs)
        
    # ----------------------------------------------------------------
    def getExtensionElements(self,asdict=True):
        '''Get all or a sole `GlobalPropertie` element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to
            `ExtensionElement` elements.  If `False`, return a list of
            all `ExtensionElement` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Extension` children

        '''
        return self.getElementsByKey(ExtensionElement,'target',asdict)

registerElement(Extension)
    
# --------------------------------------------------------------------
class ExtensionElement(Element):
    TAG    = Element.MODULE + 'ExtensionElement'
    UNIQUE = ['target']
    
    def __init__(self,
                 extension,
                 node         = None,
                 target       = ''):
        super().__init__(extension,
                         self.TAG,
                         node = node,
                         target = target)


    def getTarget(self):
        return self['target']

    @property 
    def target(self):
        return self.getTarget()
    
    def getSelect(self):
        parts = self.target.split('/')
        specs = [p.split(':') for p in parts]
        return specs


registerElement(ExtensionElement)
        
# --------------------------------------------------------------------
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . head import *
from . tile import GBXTileManager
from . piece import GBXPieceManager
from . mark import GBXMarkManager
from . board import GBXBoardManager
from . archive import Archive
from . features import Features
## END_IMPORT

# --------------------------------------------------------------------
class GBXInfo:
    def __init__(self,ar):
        '''GameBox information'''
        with VerboseGuard('Reading information') as g:
            self._bitsPerPixel = ar.word()   # 2 -> 2
            self._majorRevs    = ar.dword()  # 4 -> 6
            self._minorRevs    = ar.dword()  # 4 -> 10
            self._gameID       = ar.dword()  # 4 -> 14
            self._boxID        = ar.read(16) # 16 -> 30
            self._author       = ar.str()    # X  -> 30+X
            self._title        = ar.str()    # Y  -> 30+X+Y
            self._description  = ar.str()    # Z  -> 30+X+Y+Z
            self._password     = ar.read(16) # 16 -> 46+X+Y+Z
            self._stickyDraw   = ar.word()   # 2  -> 48+X+Y+Z
            self._compression  = ar.word()   # 2  -> 50+X+Y+Z
            self._reserved     = [ar.word() for _ in range(4)] # 4x2 -> 58+X+Y+Z
            g(f'GameBox is {self._title} by {self._author} (password: {self._password})')
        
    def __str__(self):
        return ('Information:\n'
                f'  Bits/pixel:         {self._bitsPerPixel}\n'
                f'  Major revision:     {self._majorRevs}\n'
                f'  Minor revision:     {self._minorRevs}\n'
                f'  Game ID:            {self._gameID}\n'
                f'  Box ID:             {self._boxID}\n'
                f'  Author:             {self._author}\n'
                f'  Title:              {self._title}\n'
                f'  Description:        {self._description}\n'
                f'  Password:           {self._password}\n'
                f'  Sticky Draw tools:  {self._stickyDraw}\n'
                f'  Compression level:  {self._compression}\n'
                f'  Reserved:           {self._reserved}')

    
# ====================================================================
class GameBox:
    def __init__(self,ar):
        '''Container of game'''
        with VerboseGuard(f'Reading gamebox'):
            self._header       = GBXHeader(ar,GBXHeader.BOX)
            self._info         = GBXInfo(ar)
            self._strings      = GBXStrings(ar)
            self._tileManager  = GBXTileManager(ar)
            self._boardManager = GBXBoardManager(ar)
            self._pieceManager = GBXPieceManager(ar)
            self._markManager  = GBXMarkManager(ar)

        # print(self._strings)
        # print(self._markManager)
        
    def __str__(self):
        return (str(self._header)+      
                str(self._info)+        
                str(self._strings)+     
                str(self._tileManager)+ 
                str(self._boardManager)+
                str(self._pieceManager)+
                str(self._markManager)) 

    
    @classmethod
    def fromFile(cls,filename):
        with VerboseGuard(f'Read gamebox from {filename}'):
            with Archive(filename,'rb') as ar:
                return GameBox(ar)

#
# EOF
#



## BEGIN_IMPORT
from . moduledata import ModuleData
from . withtraits import PieceSlot, DummyWithTraits
from . trait import Trait
from . traits import BasicTrait, StackTrait
from . element import DummyElement
from . mapelements import AtStart
## END_IMPORT

# ====================================================================
class SaveIO:
    '''Wrapper around a save file 
    
    Save file is
    
        "!VCSK" KEY content
    
    Key is two bytes drawn as a random number in 0-255.  Content is
    two bytes per character.  Content characters are encoded with the
    random key.
    
    Save file (.vsav) content is
    
        "begin_save" ESC
        "\" ESC
        [commands]* ESC
        "PLAYER" name password side ESC
        [map+"BoardPicker" name x y ESC]+
        "SETUP_STACK" ESC
        "TURN"+name state ESC
        "end_save"
    
    Commands are
    
        "+/" id "/" body "\"
    
    where body are
    
        "stack" "/" mapName ; x ; y ; ids "\"
        piece_type "/" piece_state   (x and y set here too) "\"
    
    x and y are pixel coordinates (sigh!).  This means we have to know
    
    - the pixel location of a hex
    - the hex coordinates of that hex
    - whether rows and columns are descending
    - if even hexes are higher or not
    
    The two first items _must_ be user supplied (I think).  When we
    add stacks or pieces, we must then get the numerical hex
    coordinates - not what the user specified in the VASSAL editor or
    the like.  Of course, this means opening the module before writing
    the patch.py script.
    
    It seems like every piece is added in a stack.
    
    The id is a numerical value.  Rather big (e.g., 1268518000806). It
    is the current number of miliseconds since epoch, with offset to
    disambiguate.
    
    The ID is the current time, taken from a milisecond clock,
    possibly adjusted up if there is a clash.  This is all managed by
    the GameState class.

    '''
    VCS_HEADER = b'!VCSK'
    VK_ESC     = chr(27)
    DEC_MAP    = {
        # 0-9
        0x30: 0x30,
        0x31: 0x30,
        0x32: 0x30,
        0x33: 0x30,
        0x34: 0x30,
        0x35: 0x30,
        0x36: 0x30,
        0x37: 0x30,
        0x38: 0x30,
        0x39: 0x30,
        # A-F
        0x41: 0x37,
        0x42: 0x37,
        0x43: 0x37,
        0x44: 0x37,
        0x45: 0x37,
        0x46: 0x37,
        # a-f
        0x61: 0x57,
        0x62: 0x57,
        0x63: 0x57,
        0x64: 0x57,
        0x65: 0x57,
        0x66: 0x57
    }
    ENC_MAP = [b'0',b'1',b'2',b'3',b'4',b'5',b'6',b'7',b'8',b'9',
               b'a',b'b',b'c',b'd',b'e',b'f']

    # ----------------------------------------------------------------
    @classmethod
    def decHex(cls,b):
        '''Decode a single char into a number

        If the encoded number is b, then the decoded number is
        
            b - off
    
        where off is an offset that depends on b
    
           off = 0x30   if 0x30 <= b <= 0x39
                 0x37   if 0x41 <= b <= 0x46
                 0x57   if 0x61 <= b <= 0x66
        '''
        return b - cls.DEC_MAP[b]
    # --------------------------------------------------------------------
    @classmethod
    def readByte(cls,inp,key):
        '''Read a single byte of information from input stream
    
        Two characters (c1 and c2) are read from input stream, and the
        decoded byte is then
    
            ((dechex(c1) << 4 | dechex(c2)) ^ key) & 0xFF
        
        Parameters
        ----------
        inp : stream
            Input to read from
        key : int
            Key to decode the input
    
        Returns
        -------
        b : int
            The read byte
        '''
        try:
            pair = inp.read(2)
        except Exception as e:
            from sys import stderr
            print(e,file=stderr)
            return None
    
        if len(pair) < 2:
            return None
    
        return ((cls.decHex(pair[0]) << 4 | cls.decHex(pair[1])) ^ key) & 0xFF
    # --------------------------------------------------------------------
    @classmethod
    def readSave(cls,file,alsometa=False):
        '''Read data from save file.  The data is read into lines
        returned as a list.

        '''
        from zipfile import ZipFile
        
        # We open the save file as a zip file 
        with ZipFile(file,'r') as z:
            # open the save file in the archive
            save = z.open('savedGame','r')
            
            # First, we check the header
            head = save.read(len(cls.VCS_HEADER))
            assert head == cls.VCS_HEADER, \
                f'Read header {head} is not {cls.VCS_HEADER}'
    
            # Then, read the key
            pair = save.read(2)
            key  = (cls.decHex(pair[0]) << 4 | cls.decHex(pair[1]))
    
            # Now read content, one byte at a time 
            content = ''
            while True:
                byte = cls.readByte(save,key)
                if byte is None:
                    break
    
                # Convert byte to character 
                content += chr(byte)
    
            lines = content.split(cls.VK_ESC)

            if alsometa:
                savedata = z.read(VSav.SAVE_DATA)
                moduledata = z.read(VMod.MODULE_DATA)

        if not alsometa:
            return key, lines

        return key,lines,savedata,moduledata

    # --------------------------------------------------------------------
    @classmethod
    def writeByte(cls,out,byte,key):
        '''Write a single byte

        Parameters
        ----------
        out : IOStream
            Stream to write to
        byte : char
            Single byte to write
        key : int
            Key to encode with (defaults to 0xAA - alternating 0's and 1's)
        '''
        b    = ord(byte) ^ key
        pair = cls.ENC_MAP[(b & 0xF0) >> 4], cls.ENC_MAP[b & 0x0F]
        out.write(pair[0])
        out.write(pair[1])

    # --------------------------------------------------------------------
    @classmethod
    def writeInZip(cls,z,key,lines,filename='savedGame'):
        '''Write a save file in a zip file (VMod)'''
        # open the save file in the archive
        with z.open(filename,'w') as save:
            # Write header
            save.write(cls.VCS_HEADER)
    
            # Split key
            pair = cls.ENC_MAP[(key & 0xF0) >> 4], cls.ENC_MAP[(key & 0x0F)]
            save.write(pair[0])
            save.write(pair[1])
    
            # Form content
            content = cls.VK_ESC.join(lines)
    
            # Write each character as two
            for c in content:
                cls.writeByte(save, c, key)
        
    # --------------------------------------------------------------------
    @classmethod
    def writeSave(cls,file,key,lines,savedata=None,moduledata=None):
        '''Write a save file'''
        from zipfile import ZipFile, ZIP_DEFLATED
        
        # We open the save file as a zip file 
        with ZipFile(file,'w',ZIP_DEFLATED) as z:
            cls.writeInZip(z,key,lines,filename='savedGame')

            if savedata is not None:
                z.writestr(VSav.SAVE_DATA,savedata)
                z.writestr(VMod.MODULE_DATA,moduledata)
        
# ====================================================================
#
# VSave file
#
class SaveFile:
    def __init__(self,game,firstid=None):
        '''Creates a save file to add positions to'''
        from time import time
        self._game     = game
        self._counters = {}
        self._stacks   = {}
        self._pieces   = self._game.getPieces(asdict=True)
        self._nextId   = (int(time()*1000) - 360000
                          if firstid is None else firstid)
        
    def add(self,grid,mapname,**kwargs):
        '''Add pieces to the save.

        Parameters
        ----------
        grid : BaseGrid
            Grid to add pieces to 
        kwargs : dict
            Either a map from piece name to hex position,
            Or a map from hex position to list of pieces
        '''
        for k,v in kwargs.items():
            # print('Add to save',k,v)
            self._add(grid,mapname,k,v)

    def addNoGrid(self,mapName,mapping):
        for k,v in mapping.items():
            # print('Add to save',k,v)
            self._add(None,mapName,k,v)
        

    def _add(self,grid,mapName,k,v):

        '''Add to the save'''
        with VerboseGuard(f'Adding piece(s) to save: {len(k)}') as vg:
            # print(f'Adding {k} -> {v}')
            loc       = None
            piece     = self._pieces.get(k,None)
            pieces    = []
            boardName = (grid.getMap()['mapName']
                         if mapName is None else mapName)
            # print(f'Map name: {mapName}')
            vg(f'Adding to {boardName}')
            if piece is not None:
                vg(f'Key {k} is a piece')
                #print(f'Key is piece: {k}->{piece}')
                pieces.append(piece)
                loc = v
            else:
                vg(f'Key {k} is a location')
                # Key is not a piece name, so a location
                loc = k
                # Convert value to iterable 
                try:
                    iter(v)
                except:
                    v = list(v)
                
                for vv in v:
                    if isinstance(vv,PieceSlot):
                        pieces.append(vv)
                        continue
                    if isinstance(vv,str):
                        piece = self._pieces.get(vv,None)
                        if piece is None:
                            continue
                        pieces.append(piece)
            
            vg(f'Loc: {loc} -> {pieces}')
            if len(pieces) < 1:
                return
            
            if (mapName,loc) not in self._stacks:
                vg(f'Adding stack {mapName},{loc}')
                coord = grid.getLocation(loc) if grid is not None else loc
                if coord is None:
                    print(f'did not get coordinates from {loc}')
                    return
                self._stacks[(mapName,loc)] = {
                    'x': coord[0],
                    'y': coord[1],
                    'pids': [] }
                    
            place = self._stacks[(mapName,loc)]
            for piece in pieces:
                name    = piece['entryName']
                gpid    = piece['gpid']
                counter = self._counters.get((name,gpid),None)
                vg(f'Got counter {counter} for {name},{gpid}')

                if counter is None:
                    if gpid == 0:
                        print(f'making new counter with pid={self._nextId}: '
                              f'{gpid}')
                        gpid = self._nextId
                        self._nextId += 1
                        
            
                    vg(f'Save adding counter with pid={gpid}')
                    counter = {'pid':   gpid,
                               'piece': piece,
                               'board': mapName,
                               'x':     place['x'],
                               'y':     place['y'],
                               }
                    self._counters[(name,gpid)] = counter
                    
                vg(f'Adding to stack {mapName},{loc}: {counter}')
                place['pids'].append(counter['pid'])

    def getLines(self,update=None):
        '''Get the final lines of code'''
        key   = 0xAA # fixed key
        
        lines = ['begin_save',
                 '',
                 '\\']

        self._pieceLines(lines,update=update)
        self._otherLines(lines)
        
        lines.append('end_save')
        return lines

    def _pieceLines(self,lines,update=lambda t:t):
        '''Add piece lines to save file

        Parameters
        ----------
        lines : list
            The lines to add
        '''
        # print(self._counters)
        for (name,gpid),counter in self._counters.items():
            iden   = counter['pid']
            piece  = counter['piece']
            traits = piece.getTraits()
            traits = Trait.flatten(traits,self._game)
            # Get last - trait (basic piece), and modify coords
            basic  = traits[-1]
            basic['map'] = counter['board']
            basic['x']   = counter['x']
            basic['y']   = counter['y']
            # Set old location if possible
            parent = piece.getParent(DummyElement,checkTag=False)
            if parent is not None and parent._node.nodeName == AtStart.TAG:
                oldLoc   = parent['location']
                oldBoard = parent['owningBoard']
                oldMap   = self._game.getBoards()[oldBoard].getMap()['mapName']
                oldX     = parent['x']
                oldY     = parent['y']
                oldZone  = None
                zones    = self._game.getBoards()[oldBoard].getZones()
                for zone in zones.values():
                    grid = zone.getGrids()[0]
                    if grid is None: continue
                    
                    coord = grid.getLocation(oldLoc)
                    if coord is None: continue

                    oldZone = zone['name']
                    oldX    = coord[0]
                    oldY    = coord[1]
                    break

                if oldZone is not None:
                    basic['properties'] = \
                        f'8;'+\
                        f'UniqueID;{iden};'+\
                        f'OldZone;{oldZone};'+\
                        f'OldLocationName;{oldLoc};'+\
                        f'OldDeckName;;'+\
                        f'OldX;{oldX};'+\
                        f'OldY;{oldY};'+\
                        f'OldBoard;{oldBoard};'+\
                        f'OldMap;{oldMap}'
                else:
                    basic['properties'] = \
                        f'7;'+\
                        f'UniqueID;{iden};'+\
                        f'OldLocationName;{oldLoc};'+\
                        f'OldDeckName;;'+\
                        f'OldX;{oldX};'+\
                        f'OldY;{oldY};'+\
                        f'OldBoard;{oldBoard};'+\
                        f'OldMap;{oldMap}'

                for trait in traits:
                    if trait.ID == TrailTrait.ID:
                        trait['map']    = oldMap
                        trait['points'] = f'1;{oldX},{oldY}'
                        trait['init']   = True

            # Let user code update the flattened traits
            if update is not None:
                update(name,traits)
            # Wrapper 
            wrap   = DummyWithTraits(self._game,traits=[])
            wrap.setTraits(*traits,iden=str(iden))
            lines.append(wrap._node.childNodes[0].nodeValue+'\\')

        layer = -1
        for key,dat in self._stacks.items():
            pids = dat.get('pids',None)
            x    = dat['x']
            y    = dat['y']
            if pids is None or len(pids) < 1:
                print(f'No pieces at {key[0]},{key[1]}')
                continue
            
            iden         =  self._nextId
            self._nextId += 1
            stack        =  StackTrait(board=key[0],x=x,y=y,pieceIds=pids,layer=layer)
            layer        = 1
            wrap         =  DummyWithTraits(self._game,traits=[])
            wrap.setTraits(stack,iden=iden)
            lines.append(wrap._node.childNodes[0].nodeValue+'\\')
            
    def _otherLines(self,lines):
        '''Add other lines to save'''
        lines.append('UNMASK\tnull')
        if self._game.getPlayerRoster():
            for r in self._game.getPlayerRoster():
                lines.extend(r.encode())
        if self._game.getNotes(single=False):
            for n in self._game.getNotes(single=False):
                lines.extend(n.encode())
        setupStack = False
        for m in self._game.getMaps(asdict=False):
            for bp in m.getBoardPicker(single=False):
                lines.extend(bp.encode())
            if not setupStack:
                atstart = m.getAtStarts(single=False)
                if atstart and len(atstart) > 0:
                    lines.append('SETUP_STACK')
                    setupStack = True
                
        # for tk,tt in self._game.getTurnTracks(asdict=True):
        #     lines.extend(tt.encode())

            
# --------------------------------------------------------------------
class SaveData(ModuleData):
    def __init__(self,root=None):
        '''Convinience wrapper'''
        super(SaveData,self).__init__(root=root)



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class RotateTrait(Trait):
    ID = 'rotate'
    def __init__(self,
                 nangles          = 6,
                 rotateCWKey      = key(']'),
                 rotateCCWKey     = key('['),
                 rotateCW         = 'Rotate CW',
                 rotateCCW        = 'Rotate CCW',
                 rotateFree       = 'Rotate ...',
                 rotateFreeKey    = key(']'),
                 rotateRndKey     = '',
                 rotateRnd        = '',
                 name             = 'Rotate',
                 description      = 'Rotate piece',
                 rotateDirectKey  = '',
                 rotateDirect     = '',
                 directExpression = '',
                 directIsFacing   = True,
                 angle            = 0):
        '''Create a Rotate trait'''
        super(RotateTrait,self).__init__()
        if nangles == 1:
            self.setType(nangles          = nangles,
                         rotateKey        = rotateFreeKey,
                         rotate           = rotateFree,
                         rotateRndKey     = rotateRndKey,
                         rotateRnd        = rotateRnd,
                         name             = name,
                         description      = description,
                         rotateDirectKey  = rotateDirectKey,
                         rotateDirect     = rotateDirect,
                         directExpression = directExpression,
                         directIsFacing   = directIsFacing)
        else:
            self.setType(nangles          = nangles,
                         rotateCWKey      = rotateCWKey,
                         rotateCCWKey     = rotateCCWKey,
                         rotateCW         = rotateCW,
                         rotateCCW        = rotateCCW,
                         rotateRndKey     = rotateRndKey,
                         rotateRnd        = rotateRnd,
                         name             = name,
                         description      = description,
                         rotateDirectKey  = rotateDirectKey,
                         rotateDirect     = rotateDirect,
                         directExpression = directExpression,
                         directIsFacing   = directIsFacing)
            
        self.setState(angle = int(angle) if nangles > 1 else float(angle))

Trait.known_traits.append(RotateTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class MarkTrait(Trait):
    ID = 'mark'
    def __init__(self,name='',value=''):
        '''Create a mark trait (static property)'''
        super(MarkTrait,self).__init__()
        self.setType(name = name)
        self.setState(value = value)


Trait.known_traits.append(MarkTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
## END_IMPORT

# --------------------------------------------------------------------
class SendtoTrait(Trait):
    ID = 'sendto'
    LOCATION = 'L'
    ZONE     = 'Z'
    REGION   = 'R'
    GRID     = 'G'
    COUNTER  = 'A'
    def __init__(self,
                 mapName     = '',
                 boardName   = '',
                 name        = '',
                 key         = key('E'),
                 restoreName = 'Restore',
                 restoreKey  = key('R'),
                 x           = 200,# Location
                 y           = 200,# Location
                 xidx        = 0,  # All - extra x
                 yidx        = 0,  # All - extra y
                 xoff        = 1,  # All - factor on xidx
                 yoff        = 1,  # All - factor on yidx
                 description = '',
                 destination = LOCATION,
                 zone        = '',  # Zone and region - expression
                 region      = '',  # Region - expression
                 expression  = '',  # Counter - expression?
                 position    = ''): # Grid - Fixed
        '''Create a send to trait (VASSAL.counter.SendToLocation)'''
        self.setType(name           = name,# NAME
                     key            = key,# KEY , MODIFIER
                     mapName        = mapName,# MAP
                     boardName      = boardName,# BOARD
                     x              = x,
                     y              = y,# X ; Y
                     restoreName    = restoreName,# BACK
                     restoreKey     = restoreKey,# KEY , MODIFIER
                     xidx           = xidx,
                     yidx           = yidx,# XIDX ; YIDX
                     xoff           = xoff,
                     yoff           = yoff,# XOFF ; YOFF
                     description    = description,# DESC
                     destination    = destination,# DEST type
                     zone           = zone,# ZONE
                     region         = region,# REGION
                     expression     = expression,# EXPRESSION
                     position       = position)                   # GRIDPOS
        self.setState(backMap = '', backX = '', backY = '')

Trait.known_traits.append(SendtoTrait)

#
# EOF
#



#
from . area import AreaTrait
from . basic import BasicTrait
from . calculatedproperty import CalculatedTrait
from . clone import CloneTrait
from . delete import DeleteTrait
from . globalcommand import GlobalCommandTrait
from . globalhotkey import GlobalHotkeyTrait
from . dynamicproperty import *
from . nostack import *
from . deselect import *
from . globalproperty import GlobalPropertyTrait
from . label import LabelTraitCodes, LabelTrait
from . layer import LayerTrait
from . mark import MarkTrait
from . mask import MaskTrait
from . moved import MovedTrait
from . place import PlaceTrait
from . prototype import PrototypeTrait
from . report import ReportTrait
from . restrictaccess import RestrictAccessTrait
from . restrictcommand import RestrictCommandsTrait
from . rotate import RotateTrait
from . sendto import SendtoTrait
from . stack import StackTrait
from . submenu import SubMenuTrait
from . trail import TrailTrait
from . trigger import TriggerTrait
from . nonrect import NonRectangleTrait
from . click import ClickTrait
from . mat import MatTrait
from . cargo import CargoTrait
from . movefixed import MoveFixedTrait
from . sheet import SheetTrait
from . hide import HideTrait
from . retrn import ReturnTrait



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class NoStackTrait(Trait):
    ID                    = 'immob'
    NORMAL_SELECT         = ''
    SHIFT_SELECT          = 'i'
    CTRL_SELECT           = 't'
    ALT_SELECT            = 'c'
    NEVER_SELECT          = 'n'
    NORMAL_BAND_SELECT    = ''
    ALT_BAND_SELECT       = 'A'
    ALT_SHIFT_BAND_SELECT = 'B'
    NEVER_BAND_SELECT     = 'Z'
    NORMAL_MOVE           = 'N'
    SELECT_MOVE           = 'I'
    NEVER_MOVE            = 'V'
    NORMAL_STACK          = 'L'
    NEVER_STACK           = 'R'
    IGNORE_GRID           = 'g'
    def __init__(self,
                 select      = NORMAL_SELECT,
                 bandSelect  = NORMAL_BAND_SELECT,
                 move        = NORMAL_MOVE,
                 canStack    = False,
                 ignoreGrid  = False,
                 description = ''):
        '''No stacking trait

        (VASSAL.counter.Immobilized)
        '''
        selectionOptions = (select +
                            (self.IGNORE_GRID if ignoreGrid else '') +
                            bandSelect)
        movementOptions  = move
        stackingOptions  = self.NORMAL_STACK if canStack else self.NEVER_STACK
                 
        '''Create a mark trait (static property)'''
        super(NoStackTrait,self).__init__()
        
        self.setType(selectionOptions = selectionOptions,
                     movementOptions  = movementOptions,
                     stackingOptions  = stackingOptions,
                     description      = description)
        self.setState()


Trait.known_traits.append(NoStackTrait)

#
# EOF
#



## BEGIN_IMPORTS
from . base import DomInspect
from . map import ZTMap
from . piece import ZTPiece
from . countersheet import ZTCounterSheet
from . dicehand import ZTDiceHand
from pywargame.common import VerboseGuard
from . scenario import ZTScenario
## END_IMPORTS

# --------------------------------------------------------------------
class ZTGameBox(DomInspect):
    
    def __init__(self,file):
        from zipfile import ZipFile
        from xml.dom.minidom import parse

        with VerboseGuard('Reading gamebox') as v:
            self._zip = ZipFile(file,'r')
            
            with self._zip.open('game-box.xml','r') as xml:
                self._gamebox =parse(xml)
                
                self._root    = self._gamebox.firstChild
                self._name    = self._get_attr(self._root,'name')
                self._desc    = self._get_attr(self._root,'description')
                self._copy    = self._get_attr(self._root,'copyright')
                self._start   = self._get_attr(self._root,'startup-scenario')
                self._icon    = self._get_attr(self._root,'icon',None)
                self._splash  = None
                self._sheets_map = {}

                if not self._icon and 'icon.bmp' in self._zip.namelist():
                    self._icon = 'icon.bmp'
                v(f'Icon is {self._icon}')
                    
                if self._icon:
                    self._splash = self._read_image(self._zip, self._icon)
                    v(f'Got splash screen {self._icon}: {self._splash}')
            
                self._dice_hands = [
                    self.parse_dice_hand(dh)
                    for dh in self._root.getElementsByTagName('dice-hand')]
                # IF no dice was defined add a 1d6
                if len(self._dice_hands) <= 0:
                    self._dice_hands = [ ZTDiceHand(None) ]

                self._maps = [
                    self.parse_map(mp)
                    for mp in self._root.getElementsByTagName('map')]

                self._scenarios = [
                    self.parse_scenario(fn)
                    for fn in self._zip.namelist() if fn.endswith('.zts')]

                # counter-sheet or terrain-sheet 
                self._counter_sheets = [
                    self.parse_counter_sheet(sh)
                    for sh in self._find_children(self._root,
                                                  'counter-sheet',
                                                  'terrain-sheet')]

                self._pieces = sum([c._piece for c in self._counter_sheets],[])
                self._cards  = sum([c._card for  c in self._counter_sheets],[])
                self.map_pieces()
        
    def parse_dice_hand(self,elem):
        return ZTDiceHand(elem)

    def parse_map(self,elem):
        m = ZTMap(elem)
        m.read_image(self._zip)
        # self._master[m._name] = m
        return m

    def parse_counter_sheet(self,elem):
        s = ZTCounterSheet(elem)
        s.read_image(self._zip)
        s.make_pieces()
        self._sheets_map[s._name] = s
        return s

    def parse_scenario(self,fn):
        with self._zip.open(fn,'r') as inp:
            return ZTScenario(inp)

    def write_image(self,dir,name,image):
        if image is None:
            return
        
        fn  = dir / f'{name}.{image.format.lower()}'
        print(f'Writing {fn}')
        image.save(filename=fn)

    def map_pieces(self):
        with VerboseGuard(f'Mapping pieces to ID and back'):
            # Cards and pieces are intermingled
            tmp = sum([c._piece+c._card for c in self._counter_sheets],[])
            self._piece_map = {
                i: p for i,p in enumerate(tmp) }
            self._piece_id = {
                v: k for k,v in self._piece_map.items()}
        
        
    def write_images(self):
        from pathlib import Path

        dir = Path('images')
        dir.mkdir(0o755,parents=True,exist_ok=True)

        for m in self._maps:
            if not m._image:
                print(f'Missing image for map {m._name}')
                continue

            self.write_image(dir,m._name,m._image._image)
            
        for i,p in enumerate(self._pieces):
            for img,nam in zip([p._front,p._back],
                           ['front','back']):

                self.write_image(dir,f'{i:04d}_{nam}',img)

        for i,p in enumerate(self._cards):
            for img,nam in zip([p._front,p._back],
                           ['front','back']):

                self.write_image(dir,f'c{i:04d}_{nam}',img)
                
        for s in self._counter_sheets:
            name = s._name
            for fac,nam in zip([s._front,s._back],
                               ['front','back']):
                if not fac:
                    continue

                self.write_image(dir,f'{name}_{nam}',fac._image)
        

    def __str__(self):
        return (f'{self._name}'+'\n'+
                f' Description: {self._desc}'+'\n'+
                f' Copyright:   {self._copy}'+'\n'+
                f' Start:       {self._start}'+'\n'+
                f' Icon:        {self._icon}'+
                ('\n' if len(self._dice_hands)>0 else '')+
                '\n'.join([str(d) for d in self._dice_hands])+
                ('\n' if len(self._maps)>0 else '')+
                '\n'.join([str(m) for m in self._maps])+
                ('\n' if len(self._counter_sheets)>0 else '')+
                '\n'.join([str(s) for s in self._counter_sheets]))
#
# EOF
#
